"""
Simple PostgreSQL Connector module.
author: Julien Tachoires <julien.tachoires@dalibo.com>
"""

import socket
import ssl
import hashlib
import struct
import re
from io import BytesIO
import datetime


class perror(Exception):
    """
    Protocol Error.
    """
    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = message


class error(Exception):
    """
    Connector Error.
    """
    def __init__(self, code, typ, message):
        Exception.__init__(self, message)
        self.code = code
        self.typ = typ
        self.message = message


def get_pgpass(pgpass=None):
    """
    Get postgres' password using pgpass file.

    http://www.postgresql.org/docs/9.2/static/libpq-pgpass.html
    http://wiki.postgresql.org/wiki/Pgpass
    """
    if pgpass is None:
        from os.path import expanduser
        home = expanduser("~")
        pgpass = "{0}/.pgpass".format(str(home))
    ret = []
    with open(pgpass, 'r') as filep:
        content = filep.readlines()
        for line in content:
            res = None
            res = re.match(r"^([^:]+):([^:]+):([^:]+):([^:]+):(.*)$", line)
            if res is not None:
                ret.append(res.group(1, 2, 3, 4, 5))
        return ret
    raise Exception("pgpass file not found")


def pg_escape(in_string, escapee_char=r"'"):
    out_string = ''
    out_string += escapee_char
    out_string += re.sub(escapee_char, escapee_char * 2, in_string)
    out_string += escapee_char
    return out_string


class protocol3(object):
    """
    PostgreSQL FE/BE protocol 3.0 implementation.
    """
    _version = 196608
    _ssl_code = 80877103

    def __init__(self,):
        # Response parsers mapping
        self._response_parsers = {
            b'R': 'parse_authentication_response',
            b'S': 'parse_parameter_status',
            b'K': 'parse_backend_key_data',
            b'Z': 'parse_ready_for_query',
            b'E': 'parse_error',
            b'T': 'parse_row_description',
            b'D': 'parse_data_row',
            b'C': 'parse_command_complete',
            b'H': 'parse_copy_out_response',
            b'd': 'parse_copy_data',
            b'c': 'parse_copy_done',
            b'N': 'parse_notice',
            b'W': 'parse_copy_both_response',
            # Not implemented
            b'A': 'ignore',  # Notification
            b'n': 'ignore',  # No Data
            b'I': 'ignore',  # Empty query
        }

    def ssl_request(self,):
        """
        SSLRequest for SSL negociation.
        """
        return struct.pack('!L', 8) + struct.pack('!L', self._ssl_code)

    def startup(self, user, database='', replication=False):
        """
        Startup packet.
        """
        data = struct.pack('!L', self._version) + b'user' + b'\x00' \
            + user.encode() + b'\x00' + b'database' + b'\x00' \
            + database.encode()
        if replication:
            data += b'\x00replication\x001'
        data += b'\x00\x00'
        return struct.pack('!L', len(data) + 4) + data

    def terminate(self,):
        """
        End of session.
        """
        return b'X' + struct.pack('!L', 4)

    def password_message(self, password):
        """
        PasswordMessage packet.
        """
        data = password.encode() + b'\x00'
        return b'p' + struct.pack('!L', len(data) + 4) + data

    def query(self, query):
        """
        Simple query mode.
        """
        data = query.encode() + b'\x00'
        return b'Q' + struct.pack('!L', len(data) + 4) + data

    def copy_data(self, sdata):
        data = b'd' + struct.pack('!L', len(sdata) + 4) + sdata
        return data

    def ignore(self, data):
        """
        Dummy parser for not implemented responses.
        """
        return (data[0:1], None)

    def _check_message_length(self, message, length):
        """
        Verify real message length.
        """
        if length != len(message) - 1:
            raise perror("Unvalid response length.")

    def parse_ssl_response(self, data):
        """
        Parse responses for SSL negotiation.
        Returns True if the postgresql backend accepts SSL.
        """
        if len(data) != 1:
            raise perror("Bad SSLRequest response")
        if data.decode() == 'S':
            return True
        return False

    def parse_message(self, data):
        """
        Wrapper around responses parsers.
        """
        if len(data) == 0:
            raise perror("Response is empty")
        return getattr(self, self._response_parsers[data[0:1]])(data)

    def parse_notice(self, data):
        """
        NoticeResponse parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        ret = []
        if length > 6:
            for raw in data[5:].split(b'\x00'):
                if len(raw) == 0:
                    continue
                code = struct.unpack('!c', raw[0:1])[0]
                string = raw[1:].decode()
                ret.append((code, string))
        return (data[0:1], ret)

    def parse_authentication_response(self, data):
        """
        Authentication response parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        code = struct.unpack('!L', data[5:9])[0]
        extra = {}
        extra['code'] = code
        if code == 5:
            extra['salt'] = struct.unpack('!4s', data[9:13])[0]
        return (data[0:1], extra)

    def parse_parameter_status(self, data):
        """
        ParameterStatus response parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        i = 0
        extra = {}
        for val in data[5:].split(b'\x00'):
            if (i % 2) == 0:
                key = val.decode()
            else:
                extra[key] = val.decode()
            i += 1
        return (data[0:1], extra)

    def parse_backend_key_data(self, data):
        """
        BackendKeyData response parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        extra = {
            'pid': struct.unpack('!L', data[5:9])[0],
            'key': struct.unpack('!L', data[9:13])[0]
        }
        return (data[0:1], extra)

    def parse_ready_for_query(self, data):
        """
        ReadyForQuery response parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        extra = {
            'status': data[5:6].decode()
        }
        return (data[0:1], extra)

    def parse_error(self, data):
        """
        Error parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        if length == 4:
            return (data[0], None)
        typ = data[5:6].decode()
        string = data[6:]
        return (data[0:1], {'type': typ, 'string': string})

    def parse_row_description(self, data):
        """
        RowDescription parser.
        """
        (length, nb_fields) = struct.unpack('!LH', data[1:7])
        self._check_message_length(data, length)
        ndata = data[7:]
        ret = []
        if nb_fields > 0:
            pos = 0
            while pos < nb_fields:
                eop = ndata.index(b'\x00')
                name = ndata[0:eop].decode()
                ndata = ndata[eop+1:]
                (table_oid, col_oid, type_oid, type_size, typmod, format_code)\
                    = struct.unpack('!LhLhlh', ndata[0:18])
                ndata = ndata[18:]
                ret.append({
                    'name': name,
                    'table_oid': table_oid,
                    'col_oid': col_oid,
                    'type_oid': type_oid,
                    'type_size': type_size,
                    'typmod': typmod,
                    'format_code': format_code
                })
                pos += 1
        return (data[0:1], ret)

    def parse_data_row(self, data):
        """
        DataRow parser.
        """
        (length, nb_fields) = struct.unpack('!LH', data[1:7])
        self._check_message_length(data, length)
        ret = []
        if nb_fields > 0:
            pos = 0
            cur = 7
            while pos < nb_fields:
                col_length = struct.unpack('!l', data[cur:cur+4])[0]
                cur += 4
                if col_length > 0:
                    value = data[cur:cur+col_length].decode()
                    cur += col_length
                else:
                    value = None
                ret.append(value)
                pos += 1
        return (data[0:1], ret)

    def parse_command_complete(self, data):
        """
        CommandComplete parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        body = data[5:].decode()
        ret = {}
        if body[0:6] == 'INSERT':
            res = re.match('^INSERT ([0-9]+) ([0-9]+)', body)
            if res is None:
                raise perror("Unvalid CommandComplete message.")
            ret = {
                'command': 'INSERT',
                'oid_table': res.group(1),
                'nb_rows': int(res.group(2))
            }
        else:
            res = re.match('^([A-Z]+) ([0-9]+)', body)
            if res is not None:
                ret = {'command': res.group(1), 'nb_rows': int(res.group(2))}
            else:
                ret = {'command': body[0:-1], 'nb_rows': 0}
        return (data[0:1], ret)

    def parse_copy_out_response(self, data):
        """
        CopyOutResponse parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        n_col = struct.unpack('!H', data[6:8])[0]
        ret = []
        pos = 0
        ndata = data[8:]
        while pos < n_col:
            ret.append(struct.unpack('!H', ndata[0:2])[0])
            ndata = ndata[2:]
            pos += 1
        return (data[0:1], ret)

    def parse_copy_data(self, data):
        """
        CopyData parser.
        """
        length = struct.unpack('!L', data[1:5])[0]
        self._check_message_length(data, length)
        body = data[5:]
        return (data[0:1], body)

    def parse_copy_done(self, data):
        return (data[0:1], None)

    def parse_copy_both_response(self, data):
        return self.parse_copy_out_response(data)

    def is_error(self, typ):
        """
        Is an error message ?
        """
        return typ == b'E'

    def is_auth_md5(self, message):
        """
        Is MD5 authentication required ?
        """
        return message[0][0:1] == b'R' and message[1]['code'] == 5

    def is_auth_cleartext(self, message):
        """
        Is cleartext authentication required ?
        """
        return message[0][0:1] == b'R' and message[1]['code'] == 3

    def is_auth_ok(self, message):
        """
        Is authentication done ?
        """
        return message[0][0:1] == b'R' and message[1]['code'] == 0

    def is_auth_supported(self, message):
        """
        Is the authentication method implemented ?
        """
        return message[0][0:1] == b'R' and (
                message[1]['code'] == 3 or
                message[1]['code'] == 5 or
                message[1]['code'] == 0)

    def is_parameter_status(self, typ):
        """
        Is a ParameterStatus message ?
        """
        return typ == b'S'

    def is_ready_for_query(self, typ):
        """
        Is a ReadyForQuery message ?
        """
        return typ == b'Z'

    def is_backend_key_data(self, typ):
        """
        Is a BackendKeyData message ?
        """
        return typ == b'K'

    def is_data_row(self, typ):
        """
        Is a DataRow message ?
        """
        return typ == b'D'

    def is_row_description(self, typ):
        """
        Is a RowDescription message ?
        """
        return typ == b'T'

    def is_command_complete(self, typ):
        """
        Is a CommandComplete message ?
        """
        return typ == b'C'

    def is_empty_query_response(self, typ):
        """
        Is an EmptyQueryResponse message ?
        """
        return typ == b'I'

    def is_copy_out_response(self, typ):
        """
        Is a CopyOutResponse message ?
        """
        return typ == b'H'

    def is_copy_data(self, typ):
        """
        Is a CopyData message ?
        """
        return typ == b'd'

    def is_copy_done(self, typ):
        """
        Is a CopyDone message ?
        """
        return typ == b'c'

    def get_salt(self, message):
        """
        Get salt value from the first authentication message
        (parsed) while dealing with MD5 method.
        """
        return message[1]['salt']

    def get_eop_tags(self,):
        """
        Return a tag list corresponding to messages:
        Error, NoData, ReadyForQuery, AuthResponse, EmptyQueryResponse
        """
        return [b'n', b'I', b'E', b'Z', b'R']


class message_buffer(object):
    """
    Message buffer class.
    """
    def __init__(self,):
        self._bio = BytesIO()
        self._pos = 0

    def write(self, data):
        """
        Write data at the end of the buffer.
        """
        self._bio.seek(0, 2)
        self._bio.write(data)

    def truncate(self,):
        """
        Truncate the buffer.
        """
        self._bio.truncate(0)
        self._bio.seek(0)
        self._pos = 0

    def get_messages(self,):
        """
        Fetch messages with a generator.
        """
        pos = 0
        # Need to know the buffer length.
        self._bio.seek(0, 2)
        buf_len = self._bio.tell()
        # If empty, nothing to do.
        if buf_len == 0:
            raise StopIteration

        self._bio.seek(pos)
        while pos < buf_len:
            # Read message header: the first 5 bytes
            header = self._bio.read(5)
            if len(header) < 5:
                raise perror("Unvalid message from buffer")
            # Get message length from the header.
            length = struct.unpack('!L', header[1:5])[0]
            pos += length + 1
            if pos > buf_len:
                raise perror("Unvalid message from buffer")
            body = self._bio.read(length - 4)
            yield header + body

    def get_messages_stream(self,):
        """
        Fetch messages with a generator and stop iteration when the last
        message is not complete. This method has to be used when we don't
        want to bufferize all messages received before parsing them.
        """
        pos_eom = 0
        # Need to know the buffer length.
        self._bio.seek(0, 2)
        buf_len = self._bio.tell()
        if buf_len == 0:
            raise StopIteration

        self._bio.seek(0)
        while pos_eom < buf_len:
            # Read message header: the first 5 bytes
            header = self._bio.read(5)
            if len(header) < 5:
                self.truncate()
                self.write(header)
                raise StopIteration
            # Get message length from the header.
            length = struct.unpack('!L', header[1:5])[0]
            pos_eom += length + 1
            if pos_eom > buf_len:
                body = self._bio.read(buf_len - self._bio.tell())
                self.truncate()
                self.write(header + body)
                raise StopIteration
            body = self._bio.read(length - 4)
            yield header + body
        if pos_eom == buf_len:
            self.truncate()

    def is_eop(self, eop_msg_tags):
        """
        Will walk through the message buffer and looking for an EOP
        (End Of Packet) message.
        """
        pos = self._pos
        self._bio.seek(0, 2)
        buf_len = self._bio.tell()
        if buf_len == 0:
            return False
        while pos < buf_len:
            self._bio.seek(pos)
            header = self._bio.read(5)
            if len(header) < 5:
                return False
            length = struct.unpack('!L', header[1:5])[0]
            pos += length + 1
            if pos > buf_len:
                return False
            else:
                self._pos = pos
            if header[0:1] in eop_msg_tags:
                return True
        return False


class connector(object):
    """
    PostgreSQL connector class.
    """

    def __init__(
            self,
            host,
            port,
            user,
            password='',
            database='',
            use_ssl=True):
        # Connection parameters.
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._database = database
        self._ssl = use_ssl
        # IPv4 host address.
        self._host_ip4 = None
        # IPv6 host address.
        self._host_ip6 = None
        # Unix socket path.
        self._host_unix = None
        # Socket.
        self._socket = None
        # protocol3 instance.
        self._protocol = protocol3()
        # message_buffer instance.
        self._message_buffer = message_buffer()
        # Size in bytes to read at each socket.recv() call.
        self._socket_read_length = 2048
        # Is the authentication phase done ?
        self._is_auth = False
        # Is the backend ready for query ?
        self._is_backend_ready = False
        # Backend key.
        self._backend_key = None
        # Backend PID.
        self._backend_pid = None
        # Some values sent by the backend.
        self._parameter_status = {}
        # Current query.
        self._query = ''
        # Rows (if any) returned.
        self._rows = []
        # Number of rows affected by the query.
        self._nb_rows = None
        # Socket default timeout.
        self._default_timeout = 60  # Seconds
        # PostgreSQL version
        self._pg_version = 0
        # replication
        self._replication = False

    def _set_ip_type(self,):
        """
        Convert the given hostname into a more convenient form.
        """
        # ip4
        if re.match(r'(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d)'
                    '(\.(?:[3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}|\d)){3}$',
                    self._host):
            self._host_ip4 = self._host
        # ip6
        elif re.match(r'^[0-9a-fA-F]+:([0-9a-fA-F]*:*){0,6}:[0-9a-fA-F]+$',
                      self._host):
            self._host_ip6 = self._host
        # unix socket
        elif re.match(r'^\/.*', self._host):
            self._host_unix = self._host
        # hostname
        else:
            try:
                for addr in socket.getaddrinfo(self._host,
                                               self._port,
                                               socket.AF_UNSPEC,
                                               socket.SOCK_STREAM):
                    if addr[0] == socket.AF_INET:
                        self._host_ip4 = addr[4][0]
                    if addr[0] == socket.AF_INET6:
                        self._host_ip6 = addr[4][0]
            except socket.gaierror:
                raise error('PGC404', 'FATAL', "Unknown name or service"
                            " '{host}'".format(host=self._host))

    def _socket_read(self,):
        """
        Read (from the socket), write (into the buffer)
        """
        self._message_buffer.truncate()
        while not self._message_buffer.is_eop(self._protocol.get_eop_tags()):
            try:
                raw_data = self._socket.recv(self._socket_read_length)
            except socket.timeout as err:
                raise error('PGC105', 'FATAL', "Timeout")
            except socket.error as err:
                raise error('PGC106', 'FATAL', "Socket error: {msg}".format(
                                                                    msg=err))
            self._message_buffer.write(raw_data)

    def _get_messages(self, method):
        """
        Read and parse messages from the buffer.
        """
        for raw in method():
            try:
                yield self._protocol.parse_message(raw)
            except perror as err:
                raise error('PGC107', 'FATAL', "Protocol violation: "
                                               "{msg}".format(msg=err.message))

    def _check_message_error(self, message):
        """
        Raise an error if the message is an error coming from the backend.
        """
        if self._protocol.is_error(message[0]):
            code = 'PGC200'
            typ = 'FATAL'
            msg = 'PostgreSQL backend error.'
            if message[1] is not None:
                err_string = message[1]['string']
                err_split = err_string.split('\x00')
                msg = err_split[2][1:].decode()
                typ = err_split[0][0:].decode()
                code = err_split[1][1:].decode()

            raise error(code, typ, msg)

    def _create_new_socket(self,):
        """
        Create a new socket.
        """
        tmp_socket = None
        # Convert "host" parameter to a valid socket address
        self._set_ip_type()
        if self._host_unix is None \
                and self._host_ip4 is None \
                and self._host_ip6 is None:
            raise error('PGC101', 'FATAL', "Could not connect to "
                                           "{host}".format(host=self._host))

        # Try with Unix socket
        if self._host_unix is not None:
            try:
                tmp_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                tmp_socket.connect(self._host_unix + '/.s.PGSQL.'
                                   + str(self._port))
                return tmp_socket
            except socket.error:
                raise error('PGC101', 'FATAL', "Could not connect to "
                                               "{host}".format(
                                                   host=self._host_unix))
        # Try with IPV4
        if self._host_ip4 is not None:
            try:
                tmp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                tmp_socket.connect((self._host_ip4, self._port))
                return tmp_socket
            except socket.error:
                if self._host_ip6 is None:
                    raise error('PGC101', 'FATAL', "Could not connect to "
                                                   "{host}".format(
                                                        host=self._host_ip4))

        # Try with IPV6
        if self._host_ip6 is not None:
            try:
                tmp_socket = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
                tmp_socket.connect((self._host_ip6, self._port))
                return tmp_socket
            except socket.error:
                raise error('PGC101', 'FATAL', "Could not connect to "
                                               "{host}".format(
                                                   host=self._host_ip6))

    def _socket_send(self, data):
        """
        Write data through the socket.
        """
        length = self._socket.send(data)
        if length != len(data):
            raise error('PGC106', 'FATAL', "Could not send all data.")

    def _wrap_ssl_socket(self, tmp_socket,):
        """
        Wrap the socket with SSL layer.
        """
        if self._ssl is True:
            data = self._protocol.ssl_request()
            tmp_socket.send(data)
            res = tmp_socket.recv(self._socket_read_length)
            if self._protocol.is_error(res[0]):
                raise error('PGC103', 'FATAL', "SSL error")
            if self._protocol.parse_ssl_response(res):
                self._socket = ssl.wrap_socket(tmp_socket)
            else:
                self._socket = tmp_socket
        else:
            self._socket = tmp_socket

    def _set_parameter_status(self, key, value):
        """
        Write a key,value pair returned in ParameterStatus message
        into self._parameter_status dict.
        """
        self._parameter_status[key] = value
        version = ''
        if key == 'server_version':
            for num in value.split('.'):
                snum = ''
                num = re.sub(r"devel|beta|rc", "", str(num))
                num = re.sub(r"^(\d+)[^\d]*.*", r"\1", str(num))
                if int(num) < 10:
                    snum += '0'
                snum += str(num)
                version += snum
            if int(version) < 10000:
                version = int(version) * 100
            self._pg_version = int(version)

    def connect(self,):
        """
        Connect to the database.
        """
        # Create and connect a new socket
        tmp_socket = self._create_new_socket()
        # Wrap with SSL
        self._wrap_ssl_socket(tmp_socket)
        # Set socket timeout
        self.set_timeout(self._default_timeout)
        # Startup
        data = self._protocol.startup(self._user,
                                      self._database,
                                      self._replication)
        self._socket_send(data)
        # Get messages from socket output
        self._socket_read()
        messages = list(self._get_messages(self._message_buffer.get_messages))
        # Test if the first message is an error
        self._check_message_error(messages[0])
        # If auth. method not implemented, raise an error
        if not self._protocol.is_auth_supported(messages[0]):
            raise error('PGC102', 'FATAL',
                        "Authentication method not supported")

        if self._protocol.is_auth_md5(messages[0]):
            # MD5
            password = "md5" + hashlib.md5(
                                hashlib.md5(
                                    self._password.encode()
                                    + self._user.encode()
                                ).hexdigest().encode()
                                + self._protocol.get_salt(messages[0])
                            ).hexdigest()
            data = self._protocol.password_message(password)
            self._socket_send(data)
            self._socket_read()
            messages = list(self._get_messages(
                self._message_buffer.get_messages))
        elif self._protocol.is_auth_cleartext(messages[0]):
            # Cleartext
            data = self._protocol.password_message(self._password)
            self._socket_send(data)
            self._socket_read()
            messages = list(self._get_messages(
                self._message_buffer.get_messages))

        for message in messages:
            # Treat all messages.
            self._check_message_error(message)
            if self._protocol.is_auth_ok(message):
                self._is_auth = True
            elif self._protocol.is_parameter_status(message[0]):
                for key, val in message[1].items():
                    self._set_parameter_status(key, val)
            elif self._protocol.is_backend_key_data(message[0]):
                self._backend_key = message[1]['key']
                self._backend_pid = message[1]['pid']
            elif self._protocol.is_ready_for_query(message[0]):
                self._is_backend_ready = True
        # Ultimate check.
        if not self._is_auth:
            raise error('PGC103', 'FATAL', "Unable to connect.")
        if not self._is_backend_ready:
            raise error('PGC104', 'FATAL', "Backend not ready.")

    def close(self,):
        """
        Close the connection to the database.
        """
        data = self._protocol.terminate()
        self._socket_send(data)
        self._socket.close()

    def execute(self, query, parameters=None):
        """
        Execute a query and fetch the results.
        """
        if parameters:
            self._query = self._build_query(query, parameters)
        else:
            self._query = query
        self._rows = []
        self._nb_rows = None
        data = self._protocol.query(self._query)
        self._socket_send(data)
        self._socket_read()
        self.get_nb_rows()

    def _build_query(self, query, parameters):
        """
        Build a query using an input string and parameters (tuple).
        Matching pattern for replacement is '%s'.
        Supported types are: int, float, str, datetime.date, datetime.datetime.
        """
        pp = list()
        for p in parameters:
            if type(p) is str:
                p = "'%s'" % (re.sub(r"('|\\)", r"\\\1", p))
            elif type(p) in (datetime.date, datetime.datetime):
                p = "'%s'" % (p)
            elif type(p) in (int, float):
                p = str(p)
            else:
                raise Exception("Unsupported type %s." % (type(p)))
            pp.append(p)
        return query % tuple(pp)

    def begin(self,):
        """
        Start a new transaction.
        """
        self.execute("BEGIN")

    def commit(self,):
        """
        Commit current transaction.
        """
        self.execute("COMMIT")

    def rollback(self,):
        """
        Rollback current transaction.
        """
        self.execute("ROLLBACK")

    def get_rows(self,):
        """
        Get rows stored in self._rows.
        """
        row_desc = None
        for message in self._get_messages(self._message_buffer.get_messages):
            if self._protocol.is_row_description(message[0]):
                row_desc = message[1]
            elif self._protocol.is_data_row(message[0]):
                if row_desc is not None:
                    row = {}
                    i = 0
                    for value in message[1]:
                        if row_desc[i]['type_oid'] in [20, 21, 23]:
                            # Convert PG int2, int4 and int8 to python int()
                            try:
                                row[row_desc[i]['name']] = int(value)
                            except Exception:
                                row[row_desc[i]['name']] = None
                        elif row_desc[i]['type_oid'] in [700, 701, 1700]:
                            # Convert PG float4, float8 and numeric to python
                            # float()
                            try:
                                row[row_desc[i]['name']] = float(value)
                            except Exception:
                                row[row_desc[i]['name']] = None
                        elif row_desc[i]['type_oid'] == 16:
                            # Convert PG boolean to python's
                            if value == 't':
                                row[row_desc[i]['name']] = True
                            else:
                                row[row_desc[i]['name']] = False
                        else:
                            row[row_desc[i]['name']] = value
                        i += 1
                else:
                    # If don't have rows descriptions
                    # then store them as tuple.
                    row = tuple(message[1])
                yield row
            elif self._protocol.is_copy_data(message[0]):
                yield message[1]
            elif self._protocol.is_command_complete(message[0]):
                self._nb_rows = message[1]['nb_rows']
            elif self._protocol.is_empty_query_response(message[0]):
                self._nb_rows = 0
            else:
                self._check_message_error(message)

    def get_nb_rows(self,):
        """
        Get number of rows affected by the current query.
        """
        if self._nb_rows is None:
            for _ in self.get_rows():
                pass
        return self._nb_rows

    def set_timeout(self, timeout):
        """
        Set socket timeout value, if None: no timeout.
        """
        self._socket.settimeout(timeout)

    def get_timeout(self,):
        """
        Get socket timeout value.
        """
        return self._socket.gettimeout()

    def get_pg_version(self,):
        """
        Get PostgreSQL version.
        """
        return self._pg_version



"""
"Hello World" plugin intended to illustrate howto to implement a new API.
WARNING: This plugin is for developer use *ONLY* and shoudn't be loaded.
"""

import sys
from temboardagent.routing import add_route
from temboardagent.api_wrapper import (
    api_function_wrapper,
    api_function_wrapper_pg,
)
from temboardagent.tools import validate_parameters
from temboardagent.errors import HTTPError


def say_hello_world(config, http_context):
    """
    Basic "Hello world" API using HTTP method GET.

    Usage:
    $ export XSESSION=`curl -s -k -X POST --data '{"username":"<user>", "password":"<password>"}' https://localhost:2345/login | sed -E "s/^.+\"([a-f0-9]+)\".+$/\1/"`
    $ curl -s -k -H "X-Session:$XSESSION" "https://localhost:2345/hello" | python -m json.tool
    {
        "content": "Hello World."
    }
    """  # noqa
    return {"content": "Hello World."}


@add_route('GET', '/hello')
def get_hello(http_context,
              queue_in=None,
              config=None,
              sessions=None,
              commands=None):
    """
    Parameters:
        http_context: HTTP context containing HTTP paramaters and variables.
        queue_in: Task queue to schedule asynchronous job.
        config: Agent configuration.
        sessions: List of current sessions.
        commands: List of current commands (async. jobs).
    """
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                sys.modules[__name__],
                                'say_hello_world')


def say_hello_world_time(conn, config, http_context):
    """
    "Hello world" API using a PostgreSQL connection.

    Usage:
    $ export XSESSION=`curl -s -k -X POST --data '{"username":"<user>", "password":"<password>"}' https://localhost:2345/login | sed -E "s/^.+\"([a-f0-9]+)\".+$/\1/"`
    $ curl -s -k -H "X-Session:$XSESSION" "https://localhost:2345/hello/time" | python -m json.tool
    {
        "message": "Hello World",
        "time": "2016-09-29 10:19:37.059801+02"
    }
    """  # noqa
    conn.execute("""
SELECT 'Hello World' AS message, NOW() AS time
    """)
    row = list(conn.get_rows())[0]
    return {"message": row['message'], "time": row['time']}


@add_route('GET', '/hello/time')
def get_hello_time(http_context,
                   queue_in=None,
                   config=None,
                   sessions=None,
                   commands=None):
    return api_function_wrapper_pg(config,
                                   http_context,
                                   sessions,
                                   sys.modules[__name__],
                                   'say_hello_world_time')


# Defining a new type to validate 'something'.
T_SOMETHING = r'(^[a-z]{1,100}$)'


def say_hello_something(config, http_context):
    """
    "Hello <something>" using slug

    Usage:
    $ export XSESSION=`curl -s -k -X POST --data '{"username":"<user>", "password":"<password>"}' https://localhost:2345/login | sed -E "s/^.+\"([a-f0-9]+)\".+$/\1/"`
    $ curl -s -k -H "X-Session:$XSESSION" "https://localhost:2345/hello/toto" | python -m json.tool
    {
        "content": "Hello toto"
    }
    """  # noqa
    return {"content": "Hello %s" % (http_context['urlvars'][0])}


@add_route('GET', '/hello/'+T_SOMETHING)
def get_hello_something(http_context,
                        queue_in=None,
                        config=None,
                        sessions=None,
                        commands=None):
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                sys.modules[__name__],
                                'say_hello_something')


def say_hello_something2(config, http_context):
    """
    "Hello <something>" using GET variable.

    Usage:
    $ export XSESSION=`curl -s -k -X POST --data '{"username":"<user>", "password":"<password>"}' https://localhost:2345/login | sed -E "s/^.+\"([a-f0-9]+)\".+$/\1/"`
    $ curl -s -k -H "X-Session:$XSESSION" "https://localhost:2345/hello2/say?something=toto" | python -m json.tool
    {
        "content": "Hello toto"
    }
    """  # noqa
    if http_context and 'something' in http_context['query']:
        validate_parameters(http_context['query'], [
            ('something', T_SOMETHING, True)
        ])
        something = http_context['query']['something'][0]
        return {"content": "Hello %s" % (something)}
    else:
        raise HTTPError(444, "Parameter 'something' not sent.")


@add_route('GET', '/hello2/say')
def get_hello_something2(http_context,
                         queue_in=None,
                         config=None,
                         sessions=None,
                         commands=None):
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                sys.modules[__name__],
                                'say_hello_something2')


def say_hello_something3(config, http_context):
    """
    "Hello <something>" using POST variable.

    Usage:
    $ export XSESSION=`curl -s -k -X POST --data '{"username":"<user>", "password":"<password>"}' https://localhost:2345/login | sed -E "s/^.+\"([a-f0-9]+)\".+$/\1/"`
    $ curl -s -k -H "X-Session:$XSESSION" -X POST --data '{"something": "toto"}' "https://localhost:2345/hello3/say" | python -m json.tool
    {
        "content": "Hello toto"
    }
    """  # noqa
    if http_context and 'something' in http_context['post']:
        validate_parameters(http_context['post'], [
            ('something', T_SOMETHING, True)
        ])
        something = http_context['post']['something']
        return {"content": "Hello %s" % (something)}
    else:
        raise HTTPError(444, "Parameter 'something' not sent.")


@add_route('POST', '/hello3/say')
def get_hello_something3(http_context,
                         queue_in=None,
                         config=None,
                         sessions=None,
                         commands=None):
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                sys.modules[__name__],
                                'say_hello_something3')


def say_hello_something4(config, http_context):
    """
    "Hello <something>" using slug & exec_command.

    Usage:
    $ export XSESSION=`curl -s -k -X POST --data '{"username":"<user>", "password":"<password>"}' https://localhost:2345/login | sed -E "s/^.+\"([a-f0-9]+)\".+$/\1/"`
    $ curl -s -k -H "X-Session:$XSESSION" "https://localhost:2345/hello4/toto" | python -m json.tool
    {
        "content": "Hello toto"
    }
    """  # noqa
    from temboardagent.command import exec_command, oneline_cmd_to_array
    (return_code, stdout, stderr) = exec_command(
                                        oneline_cmd_to_array(
                                            "echo 'Hello %s'" %
                                            http_context['urlvars'][0]
                                            )
                                        )
    return {"content": stdout[:-1]}


@add_route('GET', '/hello4/'+T_SOMETHING)
def get_hello_something4(http_context,
                         queue_in=None,
                         config=None,
                         sessions=None,
                         commands=None):
    return api_function_wrapper(config,
                                http_context,
                                sessions,
                                sys.modules[__name__],
                                'say_hello_something4')



import time
from os import getloadavg
import re

from temboardagent.queue import Queue
from temboardagent.notification import NotificationMgmt
from temboardagent.inventory import SysInfo, PgInfo


def get_metrics(conn, config, _=None):
    dm = DashboardMetrics(conn)
    sysinfo = SysInfo()
    pginfo = PgInfo(conn)
    return {'buffers': dm.get_buffers(),
            'hitratio': dm.get_hitratio(),
            'active_backends': dm.get_active_backends(),
            'cpu': dm.get_cpu_usage(),
            'loadaverage': dm.get_load_average(),
            'memory': dm.get_memory_usage(),
            'hostname': sysinfo.hostname(config.temboard['hostname']),
            'os_version': "%s %s" % (sysinfo.os, sysinfo.os_release),
            'databases': dm.get_stat_db(),
            'pg_uptime': dm.get_pg_uptime(),
            'n_cpu': sysinfo.n_cpu(),
            'pg_version': pginfo.version()['full'],
            'pg_data': pginfo.setting('data_directory'),
            'pg_port': pginfo.setting('port'),
            'notifications': dm.get_notifications(config)}


def get_metrics_queue(config, _=None):
    q = Queue('%s/dashboard.q' % (config.temboard['home']),
              max_length=(config.plugins['dashboard']['history_length']+1),
              overflow_mode='slide')
    dm = DashboardMetrics()
    msg = q.get_last_message()
    msg['notifications'] = dm.get_notifications(config)
    return msg


def get_history_metrics_queue(config, _=None):
    q = Queue('%s/dashboard.q' % (config.temboard['home']),
              max_length=(config.plugins['dashboard']['history_length']+1),
              overflow_mode='slide')
    return q.get_content_all_messages()


def get_info(conn, config, _):
    dm = DashboardMetrics(conn)
    sysinfo = SysInfo()
    pginfo = PgInfo(conn)
    return {
            'hostname': sysinfo.hostname(config.temboard['hostname']),
            'os_version': "%s %s" % (sysinfo.os, sysinfo.os_release),
            'pg_uptime': dm.get_pg_uptime(),
            'pg_version': pginfo.version()['full'],
            'pg_data': pginfo.setting('data_directory'),
            'pg_port': pginfo.setting('port'),
    }


def get_buffers(conn, config, _):
    dm = DashboardMetrics(conn)
    return {'buffers': dm.get_buffers()}


def get_hitratio(conn, config, _):
    dm = DashboardMetrics(conn)
    return {'hitratio': dm.get_hitratio()}


def get_active_backends(conn, config, _):
    dm = DashboardMetrics(conn)
    return {'active_backends': dm.get_active_backends()}


def get_cpu_usage(config, _):
    dm = DashboardMetrics()
    return {'cpu': dm.get_cpu_usage()}


def get_loadaverage(config, _):
    dm = DashboardMetrics()
    return {'loadaverage': dm.get_load_average()}


def get_memory_usage(config, _):
    dm = DashboardMetrics()
    return {'memory': dm.get_memory_usage()}


def get_hostname(config, _):
    sysinfo = SysInfo()
    return {'hostname': sysinfo.hostname(config.temboard['hostname'])}


def get_os_version(config, _):
    sysinfo = SysInfo()
    return {'os_version': "%s %s" % (sysinfo.os, sysinfo.os_release)}


def get_databases(conn, config, _):
    dm = DashboardMetrics(conn)
    return {'databases': dm.get_stat_db()}


def get_n_cpu(config, _):
    sysinfo = SysInfo()
    return {'n_cpu': sysinfo.n_cpu()}


def get_pg_version(conn, config, _):
    pginfo = PgInfo(conn)
    return {'pg_version': pginfo.version()['full']}


def get_pg_uptime(conn, config, _):
    dm = DashboardMetrics(conn)
    return {'pg_uptime': dm.get_pg_uptime()}


def get_pg_port(conn, config, _):
    pginfo = PgInfo(conn)
    return {'pg_port': pginfo.setting('port')}


def get_pg_data(conn, config, _):
    pginfo = PgInfo(conn)
    return {'pg_data': pginfo.setting('data_directory')}


class DashboardMetrics(object):
    conn = None
    config = None
    _instance = None

    def __init__(self, conn=None):
        self.conn = conn

    def get_buffers(self,):
        current_time = time.time()
        current_buffers = self._get_current_buffers()
        return {'nb': current_buffers,
                'time': current_time}

    def get_hitratio(self,):
        query = """SELECT CASE sum(blks_hit) WHEN 0 THEN NULL ELSE
        trunc((sum(blks_hit) - sum(blks_read)) / sum(blks_hit)*100) END
        AS hitratio FROM pg_stat_database"""
        self.conn.execute(query)
        return list(self.conn.get_rows())[0]['hitratio']

    def get_active_backends(self,):
        current_time = time.time()
        current_active_backends = self._get_current_active_backends()
        return {'nb': current_active_backends,
                'time': current_time}

    def get_cpu_usage(self,):
        sysinfo = SysInfo()
        if sysinfo.os == 'Linux':
            return self._get_cpu_usage_linux()

    def get_load_average(self,):
        return getloadavg()[0]

    def get_memory_usage(self,):
        sysinfo = SysInfo()
        if sysinfo.os == 'Linux':
            return self._get_memory_usage_linux()

    def get_stat_db(self,):
        query = """
SELECT
  count(datid) as databases,
  pg_size_pretty(sum(pg_database_size(
    pg_database.datname))::bigint) as total_size,
  to_char(now(),'HH24:MI') as time,
  sum(xact_commit)::BIGINT as total_commit,
  sum(xact_rollback)::BIGINT as total_rollback
FROM
  pg_database
  JOIN pg_stat_database ON (pg_database.oid = pg_stat_database.datid)
WHERE
  datistemplate = 'f'
        """
        self.conn.execute(query)
        row = list(self.conn.get_rows())[0]
        return {'databases': row['databases'],
                'total_size': row['total_size'],
                'time': row['time'],
                'total_commit': row['total_commit'],
                'total_rollback': row['total_rollback'],
                'timestamp': time.time()}

    def get_pg_uptime(self,):
        query = "SELECT NOW() - pg_postmaster_start_time() AS uptime"
        self.conn.execute(query)
        return list(self.conn.get_rows())[0]['uptime']

    def _get_memory_usage_linux(self,):
        mem_total = 0
        mem_cached = 0
        mem_free = 0
        mem_active = 0
        pattern_line_meminfo = re.compile('^([^:]+):\s+([0-9]+) kB$')
        with open('/proc/meminfo', 'r') as fd:
            for line in fd.readlines():
                m = pattern_line_meminfo.match(line)
                if m:
                    key = m.group(1)
                    value = m.group(2)
                    if key == "MemTotal":
                        mem_total = int(value)
                    if key == "MemFree":
                        mem_free = int(value)
                    if key == "Cached":
                        mem_cached = int(value)
        if mem_total == 0:
            raise Exception("Can't parse /proc/meminfo.")
        mem_active = mem_total - mem_free - mem_cached
        return {'total': mem_total,
                'free': round(float(mem_free) / float(mem_total) * 100, 1),
                'active': round(float(mem_active) / float(mem_total) * 100, 1),
                'cached': round(float(mem_cached) / float(mem_total) * 100, 1)}

    def _get_cpu_usage_linux(self,):
        cpu_time_snap_0 = self._get_current_cpu_usage_linux()
        time.sleep(0.1)
        cpu_time_snap_1 = self._get_current_cpu_usage_linux()
        delta_time_total = 0
        delta = {}
        for k in ['time_user', 'time_system', 'time_idle', 'time_iowait',
                  'time_steal']:
            delta_time_total += cpu_time_snap_1[k] - cpu_time_snap_0[k]
            delta[k] = cpu_time_snap_1[k] - cpu_time_snap_0[k]
        return {
                'user': round(delta['time_user'] / delta_time_total * 100, 1),
                'system': round(delta['time_system'] /
                                delta_time_total * 100, 1),
                'idle': round(delta['time_idle'] / delta_time_total * 100, 1),
                'iowait': round(delta['time_iowait'] /
                                delta_time_total * 100, 1),
                'steal': round(delta['time_steal'] / delta_time_total * 100, 1)
                }

    def _get_current_cpu_usage_linux(self,):
        ret = {}
        with open('/proc/stat', 'r') as fd:
            for line in fd.readlines():
                cols = line.split()
                if len(cols) > 0 and cols[0] == 'cpu':
                    ret['time_user'] = float(cols[1]) + float(cols[2])
                    ret['time_system'] = float(cols[3]) + float(cols[6]) + \
                        float(cols[7])
                    ret['time_idle'] = float(cols[4])
                    ret['time_iowait'] = float(cols[5])
                    ret['time_steal'] = float(cols[8])
                    break
        return ret

    def _get_current_buffers(self,):
        query = "SELECT buffers_alloc FROM pg_stat_bgwriter"
        self.conn.execute(query)
        return list(self.conn.get_rows())[0]['buffers_alloc']

    def _get_current_active_backends(self,):
        if self.conn.get_pg_version() >= 90200:
            query = """
SELECT COUNT(*) AS nb FROM pg_stat_activity WHERE state != 'idle'
            """
        else:
            query = """
SELECT COUNT(*) AS nb FROM pg_stat_activity WHERE current_query != '<IDLE>'
            """
        self.conn.execute(query)
        return list(self.conn.get_rows())[0]['nb']

    def get_notifications(self, config):
        return list(NotificationMgmt.get_last_n(config, 15))



import fcntl
import os
import time
import hashlib
import json


class Message(object):

    def __init__(self, id=-1, sign=None, content=''):
        if id == -1:
            self.id = int(time.time() * 1000000)
        else:
            self.id = id
        if sign is None:
            self.sign = hashlib.md5(content.encode('utf-8')).hexdigest()[0:9]
        else:
            self.sign = sign
        self.content = content

    def serialize(self,):
        return "%s:%s:%s\n" % (str(self.id), self.sign, self.content)


class Queue(object):
    """On disk queueing system."""

    def __init__(self, file_path, max_size=-1, max_length=-1,
                 overflow_mode='slide'):
        """
        Constructor

        file_path: queue file path
        max_size: maximum queue size (bytes), default -1: no limit
        max_length: maximum queue length, default -1: no limit
        overflow_mode: behaviour to have when queue size or length exceed the
                       max value on a push(), available values are:
                       - 'slide': the oldest message is removed and the new
                       one pushed.
                       - 'drop': the new message is dropped.
        """
        self.file_path = file_path
        self.max_size = max_size
        if self.max_size == -1:
            self.max_length = max_length
        else:
            self.max_length = -1
        self.overflow_mode = overflow_mode

    def get_length(self,):
        """ Read the whole queue file and return the number of lines. """
        try:
            with open(self.file_path, 'r') as fd:
                fcntl.flock(fd, fcntl.LOCK_SH)
                n = 0
                for _ in fd.readlines():
                    n += 1
                fcntl.flock(fd, fcntl.LOCK_UN)
                return n
        except Exception:
            return 0

    def get_last_message(self):
        return list(self.get_last_n_messages(1))[0]

    def get_last_n_messages(self, n):
        """
        Generator intended to return the last n messages from the queue.
        As far as the last records are located at the end of the file, we read
        the file backwards until the number of desired lines is reached or the
        whole file has been read. -1 means no limit.
        """
        buf_size = 8192
        try:
            with open(self.file_path, 'r') as fd:
                fcntl.flock(fd, fcntl.LOCK_SH)
                segment = None
                offset = 0
                n_line = 0
                # Move to the EOF
                fd.seek(0, os.SEEK_END)
                # Get file size using tell()
                file_size = total_size = remaining_size = fd.tell()

                while (remaining_size > 0):
                    offset = min(total_size, offset + buf_size)
                    # Move pointer to the next position.
                    fd.seek(file_size - offset)
                    # Read a chunk into the buffer.
                    buffer = fd.read(min(remaining_size, buf_size))
                    remaining_size -= buf_size
                    # Split buffer content by EOL.
                    lines = buffer.split('\n')

                    if segment is not None:
                        # Case when we need to concatenate the first uncomplete
                        # line of the last loop iter. with the last one of this
                        # current iteration.
                        if buffer[-1] is not '\n':
                            lines[-1] += segment
                        else:
                            n_line += 1
                            if (n > -1 and n_line > n):
                                fcntl.flock(fd, fcntl.LOCK_UN)
                                break
                            yield json.loads(
                                    self.parse_row_message(segment).content)
                    segment = lines[0]
                    # Read each line.
                    for idx in range(len(lines) - 1, 0, -1):
                        if len(lines[idx]):
                            n_line += 1
                            if (n > -1 and n_line > n):
                                fcntl.flock(fd, fcntl.LOCK_UN)
                                break
                            yield json.loads(
                                    self.parse_row_message(lines[idx]).content)

                if segment is not None:
                    yield json.loads(self.parse_row_message(segment).content)
                fcntl.flock(fd, fcntl.LOCK_UN)
        except Exception:
            return

    def get_content_all_messages(self):
        """
        Get all messages
        """
        try:
            buffers = list()
            with open(self.file_path, 'r') as fd:
                fcntl.flock(fd, fcntl.LOCK_SH)
                for row_msg in fd.readlines():
                    try:
                        msg = self.parse_row_message(row_msg)
                        buffers.append(json.loads(msg.content))
                    except Exception:
                        pass
                fcntl.flock(fd, fcntl.LOCK_UN)
            return buffers
        except Exception:
            return

    def get_size(self):
        """ Return queue file size. """
        try:
            with open(self.file_path, 'r') as fd:
                fcntl.flock(fd, fcntl.LOCK_SH)
                size = os.fstat(fd.fileno()).st_size
                fcntl.flock(fd, fcntl.LOCK_UN)
                return size
        except Exception:
            return 0

    def push(self, message):
        """ Push a new message. """
        if self.overflow_mode == 'drop':
            if self.max_length > -1 and self.get_length() >= self.max_length:
                return
            if self.max_size > -1 and self.get_size() >= self.max_size:
                return

        with open(self.file_path, 'a') as fd:
            # Let's hold an exclusive lock.
            fcntl.flock(fd, fcntl.LOCK_EX)
            fd.write(message.serialize())
            fcntl.flock(fd, fcntl.LOCK_UN)
            fd.close()

        if self.overflow_mode == 'slide':
            if self.max_size == -1 and self.max_length > -1:
                while self.get_length() > self.max_length:
                    self.shift()
            elif self.max_size > -1 and self.max_length == -1:
                while self.get_size() > self.max_size:
                    self.shift()

    def parse_row_message(self, row_message):
        if row_message:
            (id, sign) = row_message[:26].split(":")
            content = row_message[27:]
            return Message(id, sign, content.strip())
        else:
            return

    def shift(self, delete=True, check_msg=None):
        """ Get and remove from the queue the oldest message. """
        try:
            with open(self.file_path, 'r+') as fd:
                fcntl.flock(fd, fcntl.LOCK_EX)
                i = 0
                row_message = None
                message = None
                for line in fd.readlines():
                    if i == 0:
                        row_message = line
                        message = self.parse_row_message(row_message)
                        if delete is False:
                            fcntl.flock(fd, fcntl.LOCK_UN)
                            return message
                        if type(check_msg) is Message:
                            if not (message.id == check_msg.id and
                                    message.sign == check_msg.sign):
                                fcntl.flock(fd, fcntl.LOCK_UN)
                                return message
                        # Go back to the first line.
                        fd.seek(0)
                    else:
                        fd.write(line)
                    i += 1
                fd.truncate()
                fcntl.flock(fd, fcntl.LOCK_UN)
                fd.close()
                return message
        except IOError:
            try:
                fcntl.flock(fd, fcntl.LOCK_UN)
            except Exception:
                pass
            pass


def purge_queue_dir(queue_dir, exceptions=[]):
    """
    Remove queue files
    """
    if os.path.exists(queue_dir):
        for filename in os.listdir(queue_dir):
            if filename.endswith(".q") and filename not in exceptions:
                os.remove("%s/%s" % (queue_dir, filename))



import logging
import time

from temboardagent.routing import add_route
from temboardagent.errors import (
    HTTPError,
    NotificationError,
    SharedItem_exists,
    SharedItem_no_free_slot_left,
    SharedItem_not_found,
)
from temboardagent.sharedmemory import Session
from temboardagent.types import (
    T_COMMANDID,
    T_PASSWORD,
    T_SESSIONID,
    T_USERNAME,
)
from temboardagent.tools import validate_parameters
from temboardagent.usermgmt import auth_user, gen_sessionid
from temboardagent.spc import connector, error
from temboardagent.workers import COMMAND_DONE, COMMAND_ERROR
from temboardagent.notification import NotificationMgmt, Notification
from temboardagent.inventory import SysInfo, PgInfo


logger = logging.getLogger(__name__)


def check_sessionid(http_header, sessions):
    """
    Check X-Session is valid and the session exists.
    """
    validate_parameters(http_header,
                        [('X-Session', T_SESSIONID, False)])
    try:
        session = sessions.get_by_sessionid(
                    http_header['X-Session'].encode('utf-8'))
        session.time = time.time()
        username = session.username
        sessions.update(session)
        return username
    except SharedItem_not_found:
        raise HTTPError(401, "Invalid session.")


@add_route('POST', '/login')
def login(http_context, queue_in=None, config=None, sessions=None,
          commands=None):
    """
User login

**Example request**:

.. sourcecode:: http

    POST /login HTTP/1.1
    Content-Type: application/json

    {
        "username": "alice",
        "password": "foo!!"
    }

**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:19:48 GMT
    Content-type: application/json

    {"session": "fa452548403ac53f2158a65f5eb6db9723d2b07238dd83f5b6d9ca52ce817b63"}

:reqheader Content-Type: ``application/json``
:statuscode 200: no error
:statuscode 404: invalid username or password
:statuscode 500: internal error
:statuscode 406: username or password malformed or missing

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 404 Not Found
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:20:33 GMT
    Content-type: application/json

    {"error": "Invalid username/password."}

.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:21:01 GMT
    Content-type: application/json

    {"error": "Parameter 'password' is malformed."}


    """  # noqa
    post = http_context['post']
    # Add an unconditional sleeping time to reduce brute-force risks
    time.sleep(1)

    logger.info("Authenticating user: %s" % (post['username']))
    try:
        validate_parameters(post,
                            [('username', T_USERNAME, False),
                             ('password', T_PASSWORD, False)])
        auth_user(config.temboard['users'],
                  post['username'],
                  post['password'])
    except HTTPError as e:
        logger.exception(e.message)
        logger.info("Authentication failed.")
        raise e
    try:
        session = sessions.get_by_username(post['username'])
        if not session:
            sessionid = gen_sessionid(post['username'])
            session = Session(sessionid.encode('utf-8'),
                              time.time(),
                              post['username'].encode('utf-8'))
            sessions.add(session)
        else:
            sessionid = session.sessionid
            session.time = time.time()
            sessions.update(session)
        try:
            NotificationMgmt.push(config, Notification(
                                        username=post['username'],
                                        message="Login"))
        except NotificationError as e:
            logger.exception(e.message)

    except (SharedItem_exists, SharedItem_no_free_slot_left) as e:
        logger.exception(e.message)
        raise HTTPError(500, "Internal error.")
    return {'session': sessionid}


@add_route('GET', '/logout')
def logout(http_context, queue_in=None, config=None, sessions=None,
           commands=None):
    """
User logout

**Example request**:

.. sourcecode:: http

    GET /logout HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:33:19 GMT
    Content-type: application/json

    {"logout": true}

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session ID
:statuscode 500: internal error
:statuscode 406: session ID malformed

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:36:33 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:37:23 GMT
    Content-type: application/json

    {"error": "Parameter 'X-Session' is malformed."}


    """  # noqa
    headers = http_context['headers']
    logger.info("Removing session: %s" % (headers['X-Session']))
    try:
        username = check_sessionid(headers, sessions)
    except HTTPError as e:
        logger.exception(e.message)
        logger.info("Invalid session.")
        raise e

    try:
        NotificationMgmt.push(config, Notification(
                                        username=username,
                                        message="Logout"))
    except NotificationError as e:
        logger.exception(e.message)

    try:
        sessions.delete(headers['X-Session'].encode('utf-8'))
    except (SharedItem_exists, SharedItem_no_free_slot_left) as e:
        logger.exception(e.message)
        raise HTTPError(500, "Internal error.")
    return {'logout': True}


@add_route('GET', '/discover')
def get_discover(http_contexte, queue_in=None, config=None, sessions=None,
                 commands=None):
    """
Get global informations about the environment

**Example request**:

.. sourcecode:: http

    GET /discover HTTP/1.1


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:33:19 GMT
    Content-type: application/json

    {
        "hostname": "neptune",
        "pg_data": "/var/lib/postgresql/9.4/main",
        "pg_port": 5432,
        "plugins": ["monitoring", "dashboard", "pgconf", "administration", "activity"],
        "memory_size": 8241508352,
        "pg_version": "PostgreSQL 9.4.5 on x86_64-unknown-linux-gnu, compiled by gcc (Ubuntu 4.9.2-10ubuntu13) 4.9.2, 64-bit",
        "cpu": 4
    }

:statuscode 200: no error
:statuscode 500: internal error


    """  # noqa
    conn = connector(
        host=config.postgresql['host'],
        port=config.postgresql['port'],
        user=config.postgresql['user'],
        password=config.postgresql['password'],
        database=config.postgresql['dbname']
    )
    logger.info('Starting discovery.')
    try:
        conn.connect()
        sysinfo = SysInfo()
        pginfo = PgInfo(conn)
        ret = {
            'hostname': sysinfo.hostname(config.temboard['hostname']),
            'cpu': sysinfo.n_cpu(),
            'memory_size': sysinfo.memory_size(),
            'pg_port': pginfo.setting('port'),
            'pg_version': pginfo.version()['full'],
            'pg_data': pginfo.setting('data_directory'),
            'plugins': [plugin_name for plugin_name in
                        config.temboard['plugins']]
        }
        conn.close()
        logger.info('Discovery done.')
        return ret

    except (error, Exception, HTTPError) as e:
        logger.exception(str(e))
        logger.info('Discovery failed.')
        try:
            conn.close()
        except Exception:
            pass
        if isinstance(e, HTTPError):
            raise e
        else:
            raise HTTPError(500, "Internal error.")


@add_route('GET', '/profile')
def profile(http_context, queue_in=None, config=None, sessions=None,
            commands=None):
    """
Get current username

**Example request**:

.. sourcecode:: http

    GET /profile HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:33:19 GMT
    Content-type: application/json

    {
        "username": "alice"
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session ID
:statuscode 500: internal error
:statuscode 406: session ID malformed

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:36:33 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:37:23 GMT
    Content-type: application/json

    {"error": "Parameter 'X-Session' is malformed."}


    """  # noqa
    headers = http_context['headers']
    logger.info("Get user profile.")
    try:
        check_sessionid(headers, sessions)
    except HTTPError as e:
        logger.exception(e.message)
        logger.info("Invalid session.")
        raise e
    try:
        session = sessions.get_by_sessionid(
                    headers['X-Session'].encode('utf-8'))
        logger.info("Done.")
        return {'username': session.username}
    except SharedItem_not_found as e:
        logger.exception(e.message)
        logger.info("Failed.")
        raise HTTPError(401, "Invalid session.")


@add_route('GET', '/command/'+T_COMMANDID)
def get_command(http_context, queue_in=None, config=None, sessions=None,
                commands=None):
    headers = http_context['headers']
    logger.info("Get command status.")
    try:
        check_sessionid(headers, sessions)
    except HTTPError as e:
        logger.exception(e.message)
        logger.info("Invalid session.")
        raise e
    cid = http_context['urlvars'][0]
    try:
        command = commands.get_by_commandid(cid.encode('utf-8'))
        c_time = command.time
        c_state = command.state
        c_result = command.result
        if c_state == COMMAND_DONE or c_state == COMMAND_ERROR:
            commands.delete(cid.encode('utf-8'))
        logger.info("Done.")
        return {'cid': cid,
                'time': c_time,
                'state': c_state,
                'result': c_result}
    except SharedItem_not_found as e:
        logger.exception(e.message)
        logger.info("Failed.")
        raise HTTPError(401, "Invalid command.")


@add_route('GET', '/notifications')
def notifications(http_context, queue_in=None, config=None, sessions=None,
                  commands=None):
    """
Get all notifications from the agent.

**Example request**:

.. sourcecode:: http

    GET /notifications HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:33:19 GMT
    Content-type: application/json

    [
        {"date": "2016-04-11T16:12:38", "username": "alice", "message": "Login"},
        {"date": "2016-04-11T16:02:03", "username": "alice", "message": "Login"},
        {"date": "2016-04-11T15:51:15", "username": "alice", "message": "HBA file version '2016-04-11T15:32:53' removed."},
        {"date": "2016-04-11T15:51:10", "username": "alice", "message": "HBA file version '2016-04-11T15:47:26' removed."},
        {"date": "2016-04-11T15:51:04", "username": "alice", "message": "HBA file version '2016-04-11T15:48:50' removed."},
        {"date": "2016-04-11T15:50:57", "username": "alice", "message": "PostgreSQL reload"},
        {"date": "2016-04-11T15:50:57", "username": "alice", "message": "HBA file updated"},
        {"date": "2016-04-11T15:48:50", "username": "alice", "message": "PostgreSQL reload"}
    ]

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session ID
:statuscode 500: internal error
:statuscode 406: session ID malformed

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:36:33 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 12:37:23 GMT
    Content-type: application/json

    {"error": "Parameter 'X-Session' is malformed."}


    """  # noqa
    headers = http_context['headers']
    logger.info("Get notifications.")
    try:
        check_sessionid(headers, sessions)
    except HTTPError as e:
        logger.exception(e.message)
        logger.info("Invalid session.")
        raise e

    try:
        notifications = NotificationMgmt.get_last_n(config, -1)
        logger.info("Done.")
        return list(notifications)
    except (NotificationError, Exception) as e:
        logger.exception(e.message)
        logger.info("Failed.")
        raise HTTPError(500, "Internal error.")



import logging.config

try:
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from socketserver import ThreadingMixIn
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from SocketServer import ThreadingMixIn
    from urlparse import urlparse, parse_qs
import json
import sys
from urllib import unquote_plus
import ssl

from temboardagent.routing import get_routes
from temboardagent.errors import HTTPError, ConfigurationError
from temboardagent.daemon import set_global_reload, reload_true


logger = logging.getLogger(__name__)


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    """ Handle requests in a separate thread. """


class RequestHandler(BaseHTTPRequestHandler):
    """
    HTTP request handler.
    """
    def __init__(self, cmd_queue, commands, config, sessions, *args, **kwargs):
        """
        Constructor.
        """
        # Commands queue.
        self.cmd_queue = cmd_queue
        # Commands array in shared memory.
        self.commands = commands
        # Sessions array in shared memory.
        self.sessions = sessions
        # Configuration instance.
        self.config = config
        # HTTP server version.
        self.server_version = "temboard-agent/0.0.1"
        # HTTP request method
        self.http_method = None
        # HTTP query.
        self.query = None
        # HTTP POST content in json format.
        self.post_json = None
        # Call HTTP request handler constructor.
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def do_GET(self):
        """
        Handle HTTP GET requests.
        """
        self.http_method = 'GET'
        self.response()

    def do_POST(self,):
        """
        Handle HTTP POST requests.
        """
        self.http_method = 'POST'
        self.response()

    def do_PUT(self,):
        """
        Handle HTTP PUT requests.
        """
        self.http_method = 'PUT'
        # Nothing to do for now.
        pass

    def do_DELETE(self,):
        """
        Handle HTTP DELETE requests.
        """
        self.http_method = 'DELETE'
        self.response()

    def do_OPTIONS(self,):
        self.send_response(200, "OK")
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods',
                         'POST, GET, DELETE, OPTIONS')
        self.send_header('Access-Control-Allow-Headers',
                         "X-Requested-With, X-Session, Content-Type")
        self.send_header('Access-Control-Max-Age', '1728000')
        self.end_headers()
        logger.info(self.headers.dict['origin'])

    def log_message(self, format, *args):
        """
        Overrides log_message() for HTTP requests logging with our own logger.
        """
        logger.info("client: %s request: %s" % (
                         self.address_string(),
                         format % args))

    def response(self):
        """
        In charge to call the main routing function and to return its results
        as a valid HTTP response.
        """
        try:
            (code, message) = self.route_request()
        except HTTPError as e:
            code = e.code
            message = e.message
        except Exception as e:
            # This is an unknown error. Just inform there is an internal error.
            code = 500
            message = {'error': "Internal error."}
            logger.error("Internal error: %s" % (str(e)))
        self.send_response(int(code))
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(message).encode('utf-8'))

    def route_request(self,):
        """
        Main function in charge to route the incoming HTTP request to the right
        function.
        """
        # Let's parse and prepare url, path, query etc..
        url_parsed = urlparse(self.path, 'http')
        path = url_parsed.path
        splitpath = path.split('/')
        if len(splitpath) == 1:
            raise HTTPError(404, 'Not found.')
        root = splitpath[1]
        self.query = parse_qs(url_parsed.query)
        # Loop on each defined route in the API.
        for route in get_routes():
            urlvars = []
            is_that_route = True
            # Check that HTTP method and url root are matching.
            if route['http_method'] == self.http_method and \
               route['root'] == root:
                pos = 0
                # Check each element in the path.
                for elt in splitpath[1:]:
                    try:
                        if type(route['splitpath'][pos]) is not str:
                            # Then this is a regular expression.
                            res = route['splitpath'][pos].match(elt)
                            if res is not None:
                                # If the regexp matches, we want to get the
                                #  value and append it in urlvars.
                                urlvars.append(unquote_plus(res.group(1)))
                            else:
                                is_that_route = False
                                break
                        else:
                            if route['splitpath'][pos] != elt:
                                is_that_route = False
                                break
                    except IndexError:
                        is_that_route = False
                        break
                    pos += 1
                if is_that_route:
                    if self.http_method == 'POST':
                        # TODO: raise an HTTP error if the content-length is
                        # too large.
                        try:
                            # Load POST content expecting it is in JSON format.
                            post_raw = self.rfile.read(
                                        int(self.headers['Content-Length']))
                            self.post_json = json.loads(
                                                post_raw.decode('utf-8'))
                        except Exception as e:
                            raise HTTPError(400, 'Invalid json format: %s.'
                                                 % (str(e)))
                    http_context = {
                        'headers': self.headers,
                        'query': self.query,
                        'post': self.post_json,
                        'urlvars': urlvars
                    }
                    # Call the right API function.
                    response = getattr(sys.modules[route['module']],
                                       route['function'])(
                                http_context,
                                self.cmd_queue,
                                self.config,
                                self.sessions,
                                self.commands)
                    return (200, response)

        raise HTTPError(404, 'URL not found.')


def handleRequestsUsing(commands, queue_cmd, config, sessions):
    return lambda *args: RequestHandler(queue_cmd, commands, config,
                                        sessions, *args)


def httpd_run(commands, queue_in, config, sessions):
    """
    Serve HTTP for ever and reload configuration from the conf file on SIGHUP
    signal catch.
    """
    server_address = (config.temboard['address'], config.temboard['port'])
    handler_class = handleRequestsUsing(commands, queue_in, config, sessions)
    httpd = ThreadedHTTPServer(server_address, handler_class)
    httpd.socket = ssl.wrap_socket(httpd.socket,
                                   keyfile=config.temboard['ssl_key_file'],
                                   certfile=config.temboard['ssl_cert_file'],
                                   server_side=True)
    # We need a timeout here because the code after httpd.handle_request() call
    # is written to handle configuration re-loading and needs to be run
    # periodicaly.
    httpd.timeout = 1
    while True:
        httpd.handle_request()
        if reload_true():
            # SIGHUP caught
            # Try to load configuration from the configuration file.
            try:
                logger.info("SIGHUP signal caught, trying to reload "
                            "configuration.")
                config = config.reload()
            except (ConfigurationError, ImportError) as e:
                logger.exception(str(e))
                logger.info("Keeping previous configuration.")
                config.setup_logging()
            else:
                config.setup_logging()

                # New RequestHandler using the new configuration.
                httpd.RequestHandlerClass = handleRequestsUsing(
                    commands, queue_in, config, sessions,
                )
                logger.info("Done.")

            # Reset the global var indicating a SIGHUP signal.
            set_global_reload(False)



from multiprocessing import Lock
from multiprocessing.sharedctypes import Array
from ctypes import (Structure, c_char, c_double, c_int, c_wchar,
                    c_short)
import os
import time

from temboardagent.errors import (SharedItem_not_found, SharedItem_exists,
                                  SharedItem_bad_type_size,
                                  SharedItem_no_free_slot_left,
                                  NotificationError)
from temboardagent.types import (T_COMMANDID_SIZE, T_WORKER_SIZE,
                                 T_PARAMETERS_SIZE, T_RESULT_SIZE,
                                 T_SESSIONID_SIZE, T_USERNAME_SIZE)
from temboardagent.workers import COMMAND_START, COMMAND_DONE, COMMAND_ERROR
from temboardagent.notification import NotificationMgmt, Notification


class Command(Structure):
    """
    Command structure based on ctypes for sharedctypes' Array compatibility
    which is the only way for getting an array of structurized elements in a
    shared memory segment.
    Comands are keeped in a shared memory segment because they have to be
    accessed by:
        - HTTP server process when it needs to ensure the unicity. ie: we
          don't want to get the same command running twice at the same time.
          The other case is for GET /command/<cid> API.
        - scheduler process for the cleaning task.
        - worker process itself when it needs to update command status.
    """
    _fields_ = [
        # Command ID.
        ('commandid', c_char * T_COMMANDID_SIZE),
        # Last update timestamp.
        ('time', c_double),
        # Worker's pid.
        ('pid', c_int),
        # Worker name.
        ('worker', c_char * T_WORKER_SIZE),
        # Parameters (serialized).
        ('parameters', c_wchar * T_PARAMETERS_SIZE),
        # Command state.
        ('state', c_short),
        # Command return.
        ('result', c_wchar * T_RESULT_SIZE)
    ]


class Session(Structure):
    """
    Session structure. The Session Array is shared by the HTTP server process
    and the scheduler process.
    """
    _fields_ = [
        # Session ID.
        ('sessionid', c_char * T_SESSIONID_SIZE),
        # Last update timestamp.
        ('time', c_double),
        # Username.
        ('username', c_char * T_USERNAME_SIZE)
    ]


class Commands(object):
    """
    Commands object.
    """
    def __init__(self, size=100):
        # Lock handler.
        self.lock = Lock()
        # Array size.
        self.size = size
        # Array of Command.
        self.commands = Array(Command, self.size, lock=self.lock)

    def get_by_commandid(self, commandid):
        """
        Load a Command by commandid.
        """
        for command in self.commands:
            if command.commandid == commandid:
                return command
        raise SharedItem_not_found()

    def add(self, command):
        """
        Add a new Command in the shared array.
        """
        for i in range(0, self.size):
            if self.commands[i].commandid == command.commandid:
                raise SharedItem_exists()
        for i in range(0, self.size):
            if self.commands[i].commandid == b'':
                self.commands[i] = command
                return
        raise SharedItem_no_free_slot_left()

    def update(self, command):
        """
        Modify a Command in the shared array.
        """
        for i in range(0, self.size):
            if self.commands[i].commandid == command.commandid:
                self.commands[i] = command
                return
        raise SharedItem_not_found()

    def delete(self, commandid):
        """
        Remove a Command from the shared array.
        """
        for i in range(0, self.size):
            if self.commands[i].commandid == commandid:
                self.commands[i] = Command()
                return
        raise SharedItem_not_found()

    def add_tuple(self, t_command):
        """
        Convert the incoming tuple (t_command) as a Command and add it in the
        shared array.
        """
        command = Command()
        if len(t_command) != len(command._fields_):
            raise SharedItem_bad_type_size()
        command.commandid = t_command[0][:T_COMMANDID_SIZE]
        command.time = t_command[1]
        command.pid = t_command[2]
        command.worker = t_command[3][:T_WORKER_SIZE]
        command.parameters = t_command[4][:T_PARAMETERS_SIZE]
        command.state = t_command[5]
        command.result = t_command[6][:T_RESULT_SIZE]
        self.add(command)

    def purge_expired(self, ttl, logger):
        """
        In charge of removing old Commands (when Command's last update time +
        given TTL is prior to current timestamp and the worker finished its
        work on it) from the shared array.
        """
        for i in range(0, self.size):
            if len(self.commands[i].commandid) == T_COMMANDID_SIZE and \
                (self.commands[i].time + ttl) < time.time() and \
                (self.commands[i].state == COMMAND_DONE or
                    self.commands[i].state == COMMAND_ERROR):
                logger.debug("Removing command with commandid=%s" %
                             (self.commands[i].commandid))
                # Deletion: overwrite array element by a null Command.
                self.commands[i] = Command()
            # We need to ckeck if the processes executing commands with a
            # state = COMMAND_START are stil alive and then remove the
            # command if they are not.
            if len(self.commands[i].commandid) == T_COMMANDID_SIZE and \
               self.commands[i].state == COMMAND_START and \
               self.commands[i].pid > 0:
                try:
                    os.getpgid(self.commands[i].pid)
                except OSError:
                    logger.debug("Removing command with commandid=%s." % (
                                 self.commands[i].commandid))
                    self.commands[i] = Command()

    def check_uniqueness(self, worker, parameters):
        """
        Verify that there is no other identic command in progress.
        """
        for i in range(0, self.size):
            if len(self.commands[i].commandid) == T_COMMANDID_SIZE and \
               self.commands[i].worker == worker and \
               self.commands[i].parameters == parameters and \
               self.commands[i].state != COMMAND_DONE and \
               self.commands[i].state != COMMAND_ERROR:
                raise SharedItem_exists()


class Sessions(object):
    """
    Sessions object.
    """
    def __init__(self, size=100):
        # Lock handler.
        self.lock = Lock()
        # Array size.
        self.size = size
        # Array of Session.
        self.sessions = Array(Session, self.size, lock=self.lock)

    def get_by_sessionid(self, sessionid):
        """
        Returns a Session by a sessionid.
        """
        for session in self.sessions:
            if session.sessionid == sessionid:
                return session
        raise SharedItem_not_found()

    def get_by_username(self, username):
        """
        Returns a Session by a username.
        """
        for session in self.sessions:
            if session.username == username:
                return session

    def add(self, session):
        """
        Add a new Session.
        """
        for i in range(0, self.size):
            if self.sessions[i].sessionid == session.sessionid:
                raise SharedItem_exists()
            if self.sessions[i].username == session.username:
                self.sessions[i] = Session()
        for i in range(0, self.size):
            if self.sessions[i].sessionid == b'':
                self.sessions[i] = session
                return
        raise SharedItem_no_free_slot_left()

    def update(self, session):
        """
        Modify a Session.
        """
        for i in range(0, self.size):
            if self.sessions[i].sessionid == session.sessionid:
                self.sessions[i] = session
                return
        raise SharedItem_not_found()

    def delete(self, sessionid):
        """
        Remove a Session.
        """
        for i in range(0, self.size):
            if self.sessions[i].sessionid == sessionid:
                self.sessions[i] = Session()
                return
        raise SharedItem_not_found()

    def add_tuple(self, t_session):
        """
        Convert to Session the given tuple and add it.
        """
        session = Session()
        if len(t_session) != len(session._fields_):
            raise SharedItem_bad_type_size()
        session.sessionid = t_session[0][:T_SESSIONID_SIZE]
        session.time = t_session[1]
        session.username = t_session[2][:T_USERNAME_SIZE]
        self.add(session)

    def purge_expired(self, ttl, logger, config):
        """
        Remove old Session when the Session last update time + TTL is prior to
         current timestamp.
        """
        for i in range(0, self.size):
            if len(self.sessions[i].sessionid) == T_SESSIONID_SIZE and \
               (self.sessions[i].time + ttl) < time.time():
                logger.info("Session with sessionid=%s expired." %
                            (self.sessions[i].sessionid))
                try:
                    NotificationMgmt.push(config, Notification(
                                        username=self.sessions[i].username,
                                        message="Session expired"))
                except NotificationError as e:
                    logger.error(e.message)

                self.sessions[i] = Session()



import time
import os
import sys
import re
import logging
import signal
import json
import urllib2

try:
    from configparser import NoOptionError
except ImportError:
    from ConfigParser import NoOptionError


from temboardagent.routing import add_route, add_worker
from temboardagent.configuration import (
    PluginConfiguration,
    ConfigurationError,
)
from temboardagent.sharedmemory import Command
from temboardagent.tools import hash_id
from temboardagent.errors import (
    HTTPError,
    SharedItem_exists,
    SharedItem_no_free_slot_left,
)
from temboardagent.workers import (
    COMMAND_START,
)
from temboardagent.api import check_sessionid
from temboardagent.queue import Queue, Message
from temboardagent.tools import now
from temboardagent.inventory import SysInfo

from monitoring.inventory import host_info, instance_info
from monitoring.probes import (
    load_probes,
    probe_bgwriter,
    probe_blocks,
    probe_cpu,
    probe_db_size,
    probe_filesystems_size,
    probe_loadavg,
    probe_locks,
    probe_memory,
    probe_process,
    probe_replication,
    probe_sessions,
    probe_tblspc_size,
    probe_wal_files,
    probe_xacts,
    run_probes,
)
from monitoring.output import send_output, remove_passwords

__VERSION__ = '0.0.1'

logger = logging.getLogger(__name__)


@add_route('GET', '/monitoring/probe/sessions')
def monitoring_probe_sessions(http_context, queue_in=None, config=None,
                              sessions=None, commands=None):
    """
Run ``sessions`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/sessions HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "sessions":
        [
            {
                "idle_in_xact": 0,
                "idle_in_xact_aborted": 0,
                "no_priv": 0,
                "idle": 0,
                "datetime": "2017-04-21 08:24:45.003511+02",
                "disabled": 0,
                "waiting": 0,
                "port": 5432,
                "active": 0,
                "dbname": "temboard_test",
                "fastpath": 0
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.

**Error responses**:

.. sourcecode:: http

    HTTP/1.0 401 Unauthorized
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Invalid session."}


.. sourcecode:: http

    HTTP/1.0 406 Not Acceptable
    Server: temboard-agent/0.0.1 Python/2.7.8
    Date: Wed, 22 Apr 2015 09:58:00 GMT
    Content-type: application/json

    {"error": "Parameter 'X-Session' is malformed."}


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_sessions(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/xacts')
def monitoring_probe_xacts(http_context, queue_in=None, config=None,
                           sessions=None, commands=None):
    """
Run ``xacts`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/xacts HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "xacts":
        [
            {
                "port": 5432,
                "n_commit": 0,
                "n_rollback": 0,
                "dbname": "template1",
                "datetime": "2017-04-21 08:42:12.092111+02"
            }
        ]
    }


:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_xacts(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/locks')
def monitoring_probe_locks(http_context, queue_in=None, config=None,
                           sessions=None, commands=None):
    """
Run ``locks`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/locks HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "locks":
        [
            {
                "exclusive": 0,
                "waiting_share_row_exclusive": 0,
                "waiting_share": 0,
                "row_share": 0,
                "waiting_row_exclusive": 0,
                "share_row_exclusive": 0,
                "port": 5432,
                "share": 0,
                "waiting_access_share": 0,
                "dbname": "test",
                "row_exclusive": 0,
                "share_update_exclusive": 0,
                "access_share": 0,
                "access_exclusive": 0,
                "waiting_exclusive": 0,
                "siread": 0,
                "datetime": "2017-04-21 08:55:11.768602+02",
                "waiting_share_update_exclusive": 0,
                "waiting_row_share": 0,
                "waiting_access_exclusive": 0
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_locks(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/blocks')
def monitoring_probe_blocks(http_context, queue_in=None, config=None,
                            sessions=None, commands=None):
    """
Run ``blocks`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/blocks HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "blocks":
        [
            {
                "blks_read": 382,
                "dbname": "postgres",
                "hitmiss_ratio": 99.9998294969873,
                "blks_hit": 224042580,
                "datetime": "2017-04-21 08:57:32.11277+02",
                "port": 5432
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_blocks(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/bgwriter')
def monitoring_probe_bgwriter(http_context, queue_in=None, config=None,
                              sessions=None, commands=None):
    """
Run ``bgwriter`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/bgwriter HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "bgwriter":
        [
            {
                "checkpoint_write_time": 15113301.0,
                "checkpoints_timed": 1960,
                "buffers_alloc": 29369,
                "buffers_clean": 0,
                "buffers_backend_fsync": 0,
                "checkpoint_sync_time": 177464.0,
                "checkpoints_req": 0,
                "port": 5432,
                "buffers_backend": 42258,
                "maxwritten_clean": 0,
                "datetime": "2017-04-21 08:59:20.171443+02",
                "buffers_checkpoint": 149393,
                "stats_reset": "2017-04-14 13:37:15.288701+02"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_bgwriter(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/db_size')
def monitoring_probe_db_size(http_context, queue_in=None, config=None,
                             sessions=None, commands=None):
    """
Run ``db_size`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/db_size HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "db_size":
        [
            {
                "port": 5432,
                "size": 7021060,
                "dbname": "template1",
                "datetime": "2017-04-21 09:00:47.528365+02"
            },
            {
                "port": 5432,
                "size": 7168172,
                "dbname": "postgres",
                "datetime": "2017-04-21 09:00:47.528365+02"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_db_size(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/tblspc_size')
def monitoring_probe_tblspc_size(http_context, queue_in=None, config=None,
                                 sessions=None, commands=None):
    """
Run ``tblspc_size`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/tblspc_size HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "tblspc_size":
        [
            {
                "size": 181067120,
                "port": 5432,
                "spcname": "pg_default",
                "datetime": "2017-04-21 09:13:55.196718+02"
            },
            {
                "size": 622400,
                "port": 5432,
                "spcname": "pg_global",
                "datetime": "2017-04-21 09:13:55.196718+02"
            },
            {
                "size": null,
                "port": 5432,
                "spcname": "tbs",
                "datetime": "2017-04-21 09:13:55.196718+02"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_tblspc_size(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/filesystems_size')
def monitoring_probe_filesystems_size(http_context, queue_in=None, config=None,
                                      sessions=None, commands=None):
    """
Run ``filesystems_size`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/filesystems_size HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "filesystems_size":
        [
            {
                "device": "udev",
                "total": 10485760,
                "mount_point": "/dev",
                "used": 4096,
                "datetime": "2017-04-21 07:16:25 +0000"
            },
            {
                "device": "/dev/sda4",
                "total": 21003628544,
                "mount_point": "/",
                "used": 11889070080,
                "datetime": "2017-04-21 07:16:25 +0000"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(
            probe_filesystems_size(config.plugins['monitoring']),
            config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/cpu')
def monitoring_probe_cpu(http_context, queue_in=None, config=None,
                         sessions=None, commands=None):
    """
Run ``cpu`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/cpu HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "cpu":
        [
            {
                "time_system": 140,
                "time_steal": 0,
                "time_iowait": 10,
                "datetime": "2017-04-21 08:09:27 +0000",
                "measure_interval": 27.88518500328064,
                "time_idle": 27410,
                "cpu": "cpu0",
                "time_user": 290
            },
            {
                "time_system": 110,
                "time_steal": 0,
                "time_iowait": 10,
                "datetime": "2017-04-21 08:09:27 +0000",
                "measure_interval": 27.885642051696777,
                "time_idle": 27410,
                "cpu": "cpu1",
                "time_user": 290
            },
            {
                "time_system": 170,
                "time_steal": 0,
                "time_iowait": 1390,
                "datetime": "2017-04-21 08:09:27 +0000",
                "measure_interval": 27.885895013809204,
                "time_idle": 26040,
                "cpu": "cpu2",
                "time_user": 220
            },
            {
                "time_system": 130,
                "time_steal": 0,
                "time_iowait": 20,
                "datetime": "2017-04-21 08:09:27 +0000",
                "measure_interval": 27.88606309890747,
                "time_idle": 27370,
                "cpu": "cpu3",
                "time_user": 320
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_cpu(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/process')
def monitoring_probe_process(http_context, queue_in=None, config=None,
                             sessions=None, commands=None):
    """
Run ``process`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/process HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "process":
        [
            {
                "measure_interval": 55.731096029281616,
                "procs_total": "486",
                "forks": 165,
                "procs_blocked": 0,
                "context_switches": 31453,
                "procs_running": 4,
                "datetime": "2017-04-21 08:13:56 +0000"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_process(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/memory')
def monitoring_probe_memory(http_context, queue_in=None, config=None,
                            sessions=None, commands=None):
    """
Run ``memory`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/memory HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "memory":
        [
            {
                "mem_used": 7268151296,
                "swap_used": 0,
                "swap_total": 4026527744,
                "mem_total": 8276094976,
                "mem_cached": 2464796672,
                "mem_free": 1007943680,
                "mem_buffers": 558067712,
                "datetime": "2017-04-21 08:15:06 +0000"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_memory(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/loadavg')
def monitoring_probe_loadavg(http_context, queue_in=None, config=None,
                             sessions=None, commands=None):
    """
Run ``loadavg`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/loadavg HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "loadavg":
        [
            {
                "load1": "0.07",
                "load15": "0.09",
                "load5": "0.16",
                "datetime": "2017-04-21 08:16:16 +0000"
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_loadavg(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/wal_files')
def monitoring_probe_wal_files(http_context, queue_in=None, config=None,
                               sessions=None, commands=None):
    """
Run ``wal_files`` monitoring probe.

**Example request**:

.. sourcecode:: http

    GET /monotoring/probe/wal_files HTTP/1.1
    X-Session: 3b28ed94743e3ada57b217bbf9f36c6d1eb45e669a1ab693e8ca7ac3bd070b9e


**Example response**:

.. sourcecode:: http

    HTTP/1.0 200 OK
    Server: temboard-agent/0.0.1 Python/2.7.12
    Date: Fri, 21 Apr 2017 06:24:45 GMT
    Access-Control-Allow-Origin: *
    Content-type: application/json

    {
        "wal_files":
        [
            {
                "archive_ready": 0,
                "total_size": 201326592.0,
                "written_size": 13648,
                "datetime": "2017-04-21 08:17:12 +0000",
                "measure_interval": 9.273101091384888,
                "current_location": "53/700035B0",
                "total": 12,
                "port": 5432
            }
        ]
    }

:reqheader X-Session: Session ID
:statuscode 200: no error
:statuscode 401: invalid session
:statuscode 500: internal error
:statuscode 406: header ``X-Session`` is malformed.


    """  # noqa
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_wal_files(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


@add_route('GET', '/monitoring/probe/replication')
def monitoring_probe_replication(http_context, queue_in=None, config=None,
                                 sessions=None, commands=None):
    check_sessionid(http_context['headers'], sessions)

    try:
        output = api_run_probe(probe_replication(config.plugins['monitoring']),
                               config)
        return output
    except Exception as e:
        logger.error(str(e.message))
        raise HTTPError(500, "Internal error.")


def api_run_probe(probe_instance, config):
    """
    Run a probe instance.
    """
    config.plugins['monitoring']['conninfo'] = [{
        'host': config.postgresql['host'],
        'port': config.postgresql['port'],
        'user': config.postgresql['user'],
        'database': config.postgresql['dbname'],
        'password': config.postgresql['password'],
        'dbnames': config.plugins['monitoring']['dbnames'],
        'instance': config.postgresql['instance']
    }]
    # Validate connection information from the config, and ensure
    # the instance is available
    instances = []
    sysinfo = SysInfo()
    hostname = sysinfo.hostname(config.temboard['hostname'])
    for conninfo in config.plugins['monitoring']['conninfo']:
        logging.debug("Validate connection information on instance \"%s\"",
                      conninfo['instance'])
        instances.append(instance_info(conninfo, hostname))

    # Set home path
    probe_instance.set_home(config.temboard['home'])
    # Gather the data from probes
    data = run_probes([probe_instance], instances, delta=False)
    return data


@add_worker(b'monitoring_collector')
def monitoring_collector_worker(commands, command, config):
    """
    Run probes and push collected metrics in a queue.
    """
    signal.signal(signal.SIGTERM, monitoring_worker_sigterm_handler)

    start_time = time.time() * 1000
    logger.debug("Starting with pid=%s" % (os.getpid()))
    logger.debug("commandid=%s" % (command.commandid))
    command.state = COMMAND_START
    command.time = time.time()

    try:
        command.pid = os.getpid()
        commands.update(command)
        system_info = host_info(config.temboard['hostname'])
    except (ValueError, Exception) as e:
        logger.exception(str(e))
        logger.debug("Failed.")
        sys.exit(1)

    # Load the probes to run
    try:
        probes = load_probes(config.plugins['monitoring'],
                             config.temboard['home'])
        config.plugins['monitoring']['conninfo'] = [{
            'host': config.postgresql['host'],
            'port': config.postgresql['port'],
            'user': config.postgresql['user'],
            'database': config.postgresql['dbname'],
            'password': config.postgresql['password'],
            'dbnames': config.plugins['monitoring']['dbnames'],
            'instance': config.postgresql['instance']
        }]

        # Validate connection information from the config, and ensure
        # the instance is available
        instances = []
        for conninfo in config.plugins['monitoring']['conninfo']:
            logging.debug("Validate connection information on instance \"%s\"",
                          conninfo['instance'])
            instances.append(instance_info(conninfo, system_info['hostname']))

        # Gather the data from probes
        data = run_probes(probes, instances)

        # Prepare and send output
        output = {
            'datetime': now(),
            'hostinfo': system_info,
            'instances': remove_passwords(instances),
            'data': data,
            'version': __VERSION__
        }
        logger.debug("Collected data: %s" % (output))
        q = Queue('%s/metrics.q' % (config.temboard['home']),
                  max_size=1024 * 1024 * 10, overflow_mode='slide')
        q.push(Message(content=json.dumps(output)))
    except Exception as e:
        logger.exception(str(e))
        logger.debug("Failed.")
        sys.exit(1)

    logger.debug("Duration: %s." % (str(time.time() * 1000 - start_time)))
    logger.debug("Done.")


@add_worker(b'monitoring_sender')
def monitoring_sender_worker(commands, command, config):
    signal.signal(signal.SIGTERM, monitoring_worker_sigterm_handler)
    start_time = time.time() * 1000
    logger.debug("Starting with pid=%s" % (os.getpid()))
    logger.debug("commandid=%s" % (command.commandid))
    command.state = COMMAND_START
    command.time = time.time()
    command.pid = os.getpid()
    commands.update(command)
    c = 0
    while True:
        # Let's do it smoothly..
        time.sleep(0.5)

        q = Queue('%s/metrics.q' % (config.temboard['home']),
                  max_size=1024 * 1024 * 10, overflow_mode='slide')
        msg = q.shift(delete=False)
        if msg is None:
            break
        try:
            send_output(config.plugins['monitoring']['ssl_ca_cert_file'],
                        config.plugins['monitoring']['collector_url'],
                        config.temboard['key'],
                        msg.content)
        except urllib2.HTTPError as e:
            logger.exception(str(e))
            # On an error 409 (DB Integrity) we need to remove the message.
            if int(e.code) != 409:
                logger.debug(
                    "Duration: %s." % (str(time.time() * 1000 - start_time)))
                logger.debug("Failed.")
                sys.exit(1)
        except Exception as e:
            logger.exception(str(e))
            logger.debug(
                "Duration: %s." % (str(time.time() * 1000 - start_time)))
            logger.debug("Failed.")
            sys.exit(1)

        # If everything's fine then remove current msg from the queue
        q.shift(delete=True, check_msg=msg)

        if c > 60:
            break
        c += 1
    logger.debug("Duration: %s." % (str(time.time() * 1000 - start_time)))
    logger.debug("Done.")


def scheduler(queue_in, config, commands):
    # Schedule collector worker.
    schedule_worker(queue_in, config, commands, b'monitoring_collector')
    # Schedule sender worker.
    schedule_worker(queue_in, config, commands, b'monitoring_sender')


def schedule_worker(queue_in, config, commands, worker, parameters=''):
    # Check command uniqueness.
    try:
        commands.check_uniqueness(worker, parameters)
    except SharedItem_exists:
        return

    cid = hash_id(worker)
    command = Command(cid.encode('utf-8'), time.time(), 0, worker, parameters,
                      0, u'')
    try:
        commands.add(command)
        # Put the Command in the command queue
        queue_in.put(command)
        return
    except SharedItem_no_free_slot_left:
        return


def configuration(config):
    class Configuration(PluginConfiguration):
        def __init__(self, config, *args, **kwargs):
            PluginConfiguration.__init__(self, config.configfile, *args,
                                         **kwargs)

            self.plugin_configuration = {
                'dbnames': '*',
                'scheduler_interval': 60,
                'probes': '*',
                'collector_url': os.environ.get(
                    'TEMBOARD_MONITORING_COLLECTOR_URL', None),
                'ssl_ca_cert_file': None
            }

            try:
                self.check_section(__name__)
            except ConfigurationError:
                return

            try:
                dbnames = self.get(__name__, 'dbnames')
                self.plugin_configuration['dbnames'] = re.split(r'[,\s]+',
                                                                dbnames)
            except NoOptionError:
                pass

            try:
                probes = self.get(__name__, 'probes')
                self.plugin_configuration['probes'] = re.split(r'[,\s]+',
                                                               probes)
            except NoOptionError:
                pass

            try:
                collector_url = self.get(__name__, 'collector_url')
                self.plugin_configuration['collector_url'] = collector_url
            except NoOptionError:
                pass

            try:
                if not (self.getint(__name__, 'scheduler_interval') > 0 and
                        self.getint(__name__, 'scheduler_interval') < 86400):
                    raise ValueError()
                self.plugin_configuration['scheduler_interval'] = \
                    self.getint(__name__, 'scheduler_interval')
            except ValueError:
                logger.error("%s - configuration error: 'scheduler_interval' "
                             "must be an integer between 0 and 86400 in '%s' "
                             "section in %s."
                             % (__name__, self.configfile, __name__))
            except NoOptionError:
                pass

            try:
                self.plugin_configuration['ssl_ca_cert_file'] = (
                    self.getfile(__name__, 'ssl_ca_cert_file'))
            except NoOptionError:
                pass

    c = Configuration(config)
    return c.plugin_configuration


def monitoring_worker_sigterm_handler(signum, frame):
    logging.info("monitoring_worker - SIGTERM")
    sys.exit(1)



from distutils.util import strtobool
import logging
import os
import pdb
import sys

from .errors import UserError
from .configuration import setup_logging


logger = logging.getLogger(__name__)


def define_common_arguments(parser):
    parser.add_argument(
        '-c', '--config',
        action='store', dest='temboard_configfile',
        help="Configuration file",
    )


def cli(main):
    # A decorator to add consistent CLI bootstrap
    #
    # Decorated function must accept argv and environ arguments and return an
    # exit code.
    #
    # The decorator adds basic logging setup and error management. The
    # decorated function can just raise exception and log using logging module
    # as usual.

    def cli_wrapper(argv=sys.argv[1:], environ=os.environ):
        debug = strtobool(environ.get('DEBUG', '0'))
        if debug:
            os.environ['TEMBOARD_LOGGING_LEVEL'] = b'DEBUG'

        retcode = 1
        try:
            setup_logging(level='DEBUG' if debug else 'ERROR')
            logger.debug("Starting temBoard agent.")
            retcode = main(argv, environ) or 1
        except KeyboardInterrupt:
            logger.info('Terminated.')
        except pdb.bdb.BdbQuit:
            logger.info("Graceful exit from debugger.")
        except UserError as e:
            retcode = e.retcode
            logger.critical("%s", e)
        except Exception:
            logger.exception('Unhandled error:')
            if debug:
                pdb.post_mortem(sys.exc_info()[2])
            else:
                logger.error("This is a bug!")
                logger.error(
                    "Please report traceback to "
                    "https://github.com/dalibo/temboard-agent/issues! Thanks!"
                )

        exit(retcode)
    return cli_wrapper



import logging
import os
import sys
import imp
import time

from temboardagent.spc import connector


logger = logging.getLogger(__name__)


def load_plugins_configurations(config):
    """
    Intend to load plugins and run their configuration() function.
    Plugins are defined as a module located in plugins/ directory. The list
    of plugins to load is set into temboard section of the configuration file:
        [temboard]
        plugins = [ "plugin1", "plugin2" ]
    """

    # Get this module's path.
    path = os.path.dirname(__file__)
    ret = dict()
    # PostgreSQL version
    pg_version = 0

    while pg_version == 0:
        try:
            conn = connector(
                host=config.postgresql['host'],
                port=config.postgresql['port'],
                user=config.postgresql['user'],
                password=config.postgresql['password'],
                database=config.postgresql['dbname']
            )
            """ Trying to get PostgreSQL version number. """
            conn.connect()
            pg_version = conn.get_pg_version()
            conn.close()
        except Exception as e:
            logger.exception(str(e))
            logger.error("Not able to get PostgreSQL version number.")
            try:
                conn.close()
            except Exception:
                pass

        # If we reach this point, PostgreSQL is not available, so we
        # wait 5 seconds and try again
        if pg_version == 0:
            time.sleep(5)

    # Loop through each plugin listed in the configuration file.
    for plugin_name in config.temboard['plugins']:
        logger.info("Loading plugin '%s'." % (plugin_name,))
        fp_s = None
        try:
            # Loading compat.py file
            fp_s, path_s, desc_s = imp.find_module(
                                        'compat',
                                        [path+'/plugins/'+plugin_name])
            module_compat = imp.load_module('compat',
                                            fp_s,
                                            path_s,
                                            desc_s)
            # Check modules's PG_MIN_VERSION
            try:
                if (module_compat.PG_MIN_VERSION > pg_version):
                    # Version not supported
                    logger.error("PostgreSQL version (%s) is not supported "
                                 "(min:%s)." % (pg_version,
                                                module_compat.PG_MIN_VERSION))
                    logger.info("Failed.")
                    continue
            except ValueError as e:
                # PG_MIN_VERSION not set
                pass
        except Exception as e:
            if fp_s:
                fp_s.close()
            logger.info("Not able to load the compatibility file: compat.py.")
        logger.info("Done.")
        try:
            # Locate and load the module with imp.
            fp, pathname, description = imp.find_module(plugin_name,
                                                        [path+'/plugins'])
            module = imp.load_module(plugin_name, fp, pathname, description)
            # Try to run module's configuration() function.
            logger.info("Loading plugin '%s' configuration." % (plugin_name))
            plugin_configuration = getattr(module, 'configuration')(config)
            ret.update({module.__name__: plugin_configuration})
            logger.info("Done.")
        except AttributeError as e:
            logger.warn("No configuration: %s", e)
        except Exception as e:
            if fp:
                fp.close()
            logger.exception(str(e))
            logger.info("Failed.")

    return ret


# Global var to keep a track of timestamp of the last plugin's scheduler()
# function call.
PLUGINS_LAST_SCHEDULE = {}


def exec_scheduler(queue_in, config, commands, logger):
    """
    In charge of running the scheduling function of each plugin.
    """
    global PLUGINS_LAST_SCHEDULE
    for plugin_name in config.plugins:
        first_run = False
        if not (plugin_name in sys.modules):
            # The plugin seems not to be loaded.
            continue
        if not ('scheduler_interval' in config.plugins[plugin_name]):
            # If scheduler_interval is not set, we don't want to run it.
            continue
        if not (plugin_name in PLUGINS_LAST_SCHEDULE):
            # Check if this is the first shoot.
            PLUGINS_LAST_SCHEDULE[plugin_name] = time.time()
            first_run = True
        if not first_run and (
                (time.time() - PLUGINS_LAST_SCHEDULE[plugin_name]) <
                config.plugins[plugin_name]['scheduler_interval']):
            continue
        try:
            logger.debug("Running %s.scheduler()" % (plugin_name))
            # Call plugin's scheduler() function.
            getattr(sys.modules[plugin_name], 'scheduler')(queue_in,
                                                           config,
                                                           commands)
            PLUGINS_LAST_SCHEDULE[plugin_name] = time.time()
            logger.debug("Done.")
        except AttributeError as e:
            # scheduler() function does not exist.
            logger.debug("Function does not exist.")
            pass
        except Exception as e:
            logger.exception(str(e))
            logger.debug("Failed.")



import logging.config
from multiprocessing import Process
import signal
from os import getpid

from temboardagent.routing import get_worker
from temboardagent.daemon import (set_global_workers,
                                  scheduler_sigterm_handler,
                                  scheduler_sighup_handler,
                                  set_global_reload,
                                  reload_true, worker_sigterm_handler,
                                  worker_sighup_handler)
from temboardagent.errors import ConfigurationError
from temboardagent.pluginsmgmt import exec_scheduler


logger = logging.getLogger(__name__)


def Worker(commands, command, config):
    """
    Routing function in charge of calling the right worker function.
    """
    # Add a signal handler on SIGTERM and SIGHUP signals.
    signal.signal(signal.SIGTERM, worker_sigterm_handler)
    signal.signal(signal.SIGHUP, worker_sighup_handler)

    try:
        get_worker(command.worker)(commands, command, config)
    except KeyboardInterrupt:
        return
    except (AttributeError, Exception) as e:
        logger.error(str(e))


def Scheduler(commands, queue_in, config, sessions):
    """
    Asynchronous command scheduler in charge of:
        - fetching new async command from the command queue.
        - if any new command, starting a new worker process.
        - doing maintenance tasks like sessions and commands clean-up.
        - executing function named scheduler() from each loaded plugins.
    """
    # Add a signal handler on SIGTERM and SIGHUP signals.
    signal.signal(signal.SIGTERM, scheduler_sigterm_handler)
    signal.signal(signal.SIGHUP, scheduler_sighup_handler)

    logger.debug("Starting with pid=%s", (getpid()))
    workers = []
    while True:
        if reload_true():
            # SIGHUP signal caught.
            try:
                logger.info("SIGHUP signal caught, trying to reload"
                            " configuration.")
                config = config.reload()
                logger.info("New configuration loaded.")
            except (ConfigurationError, ImportError) as e:
                logger.exception(str(e))
                logger.info("Some error occured, keeping old configuration.")
            else:
                config.setup_logging()

            set_global_reload(False)

        try:
            # Fetch one new input from the command queue.
            # There is a 0.5s second timeout and the call isn't blocking.
            command = queue_in.get(True, 0.5)
        except KeyboardInterrupt:
            break
        except Exception:
            # No new command.
            pass
        else:
            # Start the worker process.
            newworker = Process(target=Worker,
                                args=(commands, command, config))
            workers.append(newworker)
            newworker.start()

        # Check that worker processes are still alive.
        for worker in workers:
            if not worker.is_alive():
                worker.join()
                workers.remove(worker)

        # Let store workers in a global var.
        set_global_workers(workers)
        # Execute plugins scheduler() function.
        exec_scheduler(queue_in, config, commands, logger)
        # Purge expired sessions if any.
        sessions.purge_expired(3600, logger, config)
        # Remove old unchecked commands.
        commands.purge_expired(60, logger)



from setuptools import setup, find_packages

SETUP_KWARGS = dict(
    name='temboard-agent',
    version='1.1',
    author='Dalibo',
    author_email='contact@dalibo.com',
    license='PostgreSQL',
    url='http://temboard.io/',
    description='Administration & monitoring PostgreSQL agent.',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: No Input/Output (Daemon)",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: System :: Monitoring",
    ],
    scripts=[
        'temboard-agent',
        'temboard-agent-adduser',
        'temboard-agent-password',
        'temboard-agent-register',
    ],
    data_files=[
        ('share/temboard-agent/', [
            'share/temboard-agent.conf',
            'share/temboard-agent.logrotate',
        ]),
        ('share/temboard-agent/quickstart/', [
            'share/temboard-agent.conf',
            'share/temboard-agent_CHANGEME.pem',
            'share/temboard-agent_CHANGEME.key',
            'share/temboard-agent_ca_certs_CHANGEME.pem',
            'share/temboard-agent.logrotate',
            'share/users',
        ]),
        ('lib/systemd/system', ['packaging/temboard-agent.service']),
    ],
)

if __name__ == '__main__':
    setup(**dict(
        SETUP_KWARGS,
        packages=find_packages(),
        long_description=open('README.rst').read(),
    ))



import httplib
import urllib2
import ssl
import socket
import json


class VerifiedHTTPSConnection(httplib.HTTPSConnection):
    """
    HTTPS connection class using custom SSL ca cert file
    """
    def connect(self):
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        self.sock = ssl.wrap_socket(sock,
                                    self.key_file,
                                    self.cert_file,
                                    cert_reqs=ssl.CERT_REQUIRED,
                                    ca_certs=CA_CERT_FILE)


class VerifiedHTTPSHandler(urllib2.HTTPSHandler):
    """
    HTTPS connection handler overriding https_open() method using
    VerifiedHTTPSConnection
    """
    def __init__(self, connection_class=VerifiedHTTPSConnection):
        self.specialized_conn_class = connection_class
        urllib2.HTTPSHandler.__init__(self)

    def https_open(self, req):
        return self.do_open(self.specialized_conn_class, req)


class UnverifiedHTTPSConnection(httplib.HTTPSConnection):
    """
    HTTPS connection class, without any SSL cert. check.
    """
    def connect(self):
        sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        self.sock = ssl.wrap_socket(sock)


class UnverifiedHTTPSHandler(urllib2.HTTPSHandler):
    """
    HTTPS connection handler
    """
    def __init__(self, connection_class=UnverifiedHTTPSConnection):
        self.specialized_conn_class = connection_class
        urllib2.HTTPSHandler.__init__(self)

    def https_open(self, req):
        return self.do_open(self.specialized_conn_class, req)


class RequestWithMethod(urllib2.Request):
    """
    Override urllib2.Request by adding HTTP method support.
    """
    def __init__(self, *args, **kwargs):
        self._method = kwargs.pop('method', None)
        urllib2.Request.__init__(self, *args, **kwargs)

    def get_method(self):
        return self._method if self._method else \
                super(RequestWithMethod, self).get_method()


def https_request(in_ca_cert_file, method, url, headers=None, data=None):
    """
    Send an HTTPS request
    """
    if in_ca_cert_file:
        global CA_CERT_FILE
        CA_CERT_FILE = in_ca_cert_file
        https_handler = VerifiedHTTPSHandler()
    else:
        https_handler = UnverifiedHTTPSHandler()
    url_opener = urllib2.build_opener(https_handler)
    headers_list = []
    for key, val in headers.iteritems():
        headers_list.append((key, val))
    url_opener.addheaders = headers_list
    if data:
        request = RequestWithMethod(url, data=json.dumps(data), method=method)
    else:
        request = RequestWithMethod(url, method=method)
    handle = url_opener.open(request)
    response = handle.read()
    if 'Set-Cookie' in handle.info():
        cookies = handle.info()['Set-Cookie']
    else:
        cookies = None
    handle.close()
    return (handle.code, response, cookies)



import re
from pgconf.filemgmt import (
    ConfigurationFileManager,
    read_file_content,
    save_file_content,
)
from temboardagent.errors import HTTPError


class HBAManager(ConfigurationFileManager):

    @classmethod
    def get_entries(self, filepath, version=None):
        entries = []
        for raw_line in read_file_content(filepath, version).split('\n'):
            entry = HBAEntry.parse(raw_line)
            if entry is not None:
                entries.append(entry)
        return entries

    @classmethod
    def save_entries(self, filepath, hba_entries, new_version=False):
        filecontent = ''
        for hba_entry in hba_entries:
            filecontent += str(hba_entry)
        return save_file_content(filecontent, filepath, new_version)

    @classmethod
    def options(self, conn):
        connections = ["local", "host", "hostssl", "hostnossl"]
        auth_methods = ["trust", "reject", "md5", "password", "gss", "sspi",
                        "ident", "peer", "ldap", "radius", "cert", "pam"]
        databases = ["all", "sameuser", "samerole", "replication"]
        users = ["all"]
        # Get users
        conn.execute("SELECT usename FROM pg_user ORDER BY usename ASC")
        for row in conn.get_rows():
            users.append(row['usename'])
        # Get groups containgin users
        conn.execute("""
            SELECT rolname
            FROM pg_user
            JOIN pg_auth_members
            ON (pg_user.usesysid=pg_auth_members.member)
            JOIN pg_roles
            ON (pg_roles.oid=pg_auth_members.roleid) ORDER BY rolname ASC
        """)
        for row in conn.get_rows():
            users.append("+%s" % (row['rolname']))
        # Get databases name
        conn.execute("""
            SELECT datname
            FROM pg_database
            WHERE NOT datistemplate
            ORDER BY datname ASC
        """)
        for row in conn.get_rows():
            databases.append(row['datname'])
        return {'connections': connections,
                'databases': databases,
                'users': users,
                'auth_methods': sorted(auth_methods)}


class HBAComment:

    def __init__(self, comment=''):
        self.comment = comment

    def __repr__(self):
        return "# %s\r\n" % (self.comment)


class HBAEntry:

    def __init__(self,
                 connection='',
                 database='',
                 user='',
                 address='',
                 auth_method='',
                 auth_options=''):
        self.connection = connection
        self.database = database
        self.user = user
        self.address = address
        self.auth_method = auth_method
        self.auth_options = auth_options

    def lazy_check(self):
        if self.connection not in ['host', 'hostssl', 'hostnossl', 'local']:
            raise HTTPError(406, "Invalid connection: %s" % (self.connection))
        if len(self.database) < 1:
            raise HTTPError(406, "Invalid database: %s" % (self.database))
        if len(self.user) < 1:
            raise HTTPError(406, "Invalid user: %s" % (self.user))
        if self.connection != 'local' and \
           len(self.address) == 0:
            raise HTTPError(406, "An address is required for method '%s'." %
                                 (self.connection))
        if len(self.auth_method) == 0:
            raise HTTPError(406, "Authentication method must be set.")

    def __repr__(self):
        row = "%s  %s  %s" % (
                self.connection,
                self.database,
                self.user)
        if self.connection is not 'local':
            row += " %s" % (self.address)
        row += " %s %s\r\n" % (self.auth_method, self.auth_options)
        return row

    @classmethod
    def parse(cls, raw):
        raw = raw.strip()
        if raw.startswith('#'):
            return HBAComment(raw.strip()[1:])
        if not raw.startswith('#') and len(raw) > 0:
            p_anything = r'([^\s]+(?<!\s))'
            p_anything_end = r'(\s*[^#]+)?(.*)$'
            p_connection_host = r'(host|hostssl|hostnossl)'
            p_connection_local = r'(local)'
            p_address = r'([0-9a-fA-F\.:]+/[0-9]+(?<!\s))'
            p_ip_address = r'([0-9a-fA-F\.:]+(?<!\s))'
            p_ip_mask = r'([0-9a-fA-F\.:]+(?<!\s))'

            r_host_address = re.compile(
                r'^'
                + p_connection_host + r'\s+'
                + p_anything + r'\s+'
                + p_anything + r'\s+'
                + p_address + r'\s+'
                + p_anything
                + p_anything_end)
            r_host_ip_address = re.compile(
                r'^'
                + p_connection_host + r'\s+'
                + p_anything + r'\s+'
                + p_anything + r'\s+'
                + p_ip_address + r'\s+'
                + p_ip_mask + r'\s+'
                + p_anything
                + p_anything_end)
            r_local = re.compile(
                r'^'
                + p_connection_local + r'\s+'
                + p_anything + r'\s+'
                + p_anything + r'\s+'
                + p_anything
                + p_anything_end)

            m_host = r_host_address.match(raw.strip())
            if m_host:
                return cls(
                    m_host.group(1),
                    m_host.group(2),
                    m_host.group(3),
                    m_host.group(4),
                    m_host.group(5),
                    m_host.group(6).strip()
                    if m_host.group(6)
                    and not m_host.group(6).strip().startswith('#')
                    else '')

            m_host = r_host_ip_address.match(raw.strip())
            if m_host:
                return cls(
                    m_host.group(1),
                    m_host.group(2),
                    m_host.group(3),
                    "%s %s" % (m_host.group(4), m_host.group(5)),
                    m_host.group(6),
                    m_host.group(7).strip()
                    if m_host.group(7)
                    and not m_host.group(7).strip().startswith('#')
                    else '')

            m_local = r_local.match(raw.strip())
            if m_local:
                return cls(
                    m_local.group(1),
                    m_local.group(2),
                    m_local.group(3),
                    '',
                    m_local.group(4),
                    m_local.group(5).strip()
                    if m_local.group(5)
                    and not m_local.group(5).strip().startswith('#')
                    else '')
